import pandas as pd
import pyarrow as pa
import pytest

from plateau.utils.converters import (
    converter_str,
    converter_str_set,
    converter_str_set_optional,
    converter_str_tupleset,
    converter_tuple,
    get_str_to_python_converter,
)


@pytest.mark.parametrize(
    "param,expected",
    [
        (
            # param
            "foo",
            # expected
            {"foo"},
        ),
        (
            # param
            "foo",
            # expected
            {"foo"},
        ),
        (
            # param
            b"foo",
            # expected
            {"foo"},
        ),
        (
            # param
            {"foo", "bar"},
            # expected
            {"foo", "bar"},
        ),
        (
            # param
            {"foo", b"foo"},
            # expected
            {"foo"},
        ),
        (
            # param
            ["foo", b"foo"],
            # expected
            {"foo"},
        ),
        (
            # param
            None,
            # expected
            set(),
        ),
    ],
)
def test_str_set(param, expected):
    actual = converter_str_set(param)
    assert isinstance(actual, frozenset)
    assert actual == expected
    assert all(isinstance(x, str) for x in actual)


@pytest.mark.parametrize(
    "param,expected",
    [
        (
            # param
            "foo",
            # expected
            {"foo"},
        ),
        (
            # param
            "foo",
            # expected
            {"foo"},
        ),
        (
            # param
            b"foo",
            # expected
            {"foo"},
        ),
        (
            # param
            {"foo", "bar"},
            # expected
            {"foo", "bar"},
        ),
        (
            # param
            {"foo", b"foo"},
            # expected
            {"foo"},
        ),
        (
            # param
            ["foo", b"foo"],
            # expected
            {"foo"},
        ),
        (
            # param
            None,
            # expected
            None,
        ),
    ],
)
def test_str_set_optional(param, expected):
    actual = converter_str_set_optional(param)
    assert actual == expected
    if actual is not None:
        assert isinstance(actual, frozenset)
        assert all(isinstance(x, str) for x in actual)


@pytest.mark.parametrize(
    "param,expected",
    [
        (
            # param
            "foo",
            # expected
            ("foo",),
        ),
        (
            # param
            "foo",
            # expected
            ("foo",),
        ),
        (
            # param
            b"foo",
            # expected
            ("foo",),
        ),
        (
            # param
            ["foo", "bar"],
            # expected
            ("foo", "bar"),
        ),
        (
            # param
            None,
            # expected
            (),
        ),
    ],
)
def test_str_tupleset_ok(param, expected):
    actual = converter_str_tupleset(param)
    assert isinstance(actual, tuple)
    assert actual == expected
    assert all(isinstance(x, str) for x in actual)


def test_str_tupleset_fail_duplicates():
    with pytest.raises(ValueError, match="Tuple-set contains duplicates: foo, foo"):
        converter_str_tupleset([b"foo", "foo"])


@pytest.mark.parametrize(
    "param",
    [
        # set
        {"foo", "bar"},
        # dict
        {"foo": 1, "bar": 2},
        # frozenset
        frozenset(("foo", "bar")),
    ],
)
def test_str_tupleset_fail_unstable(param):
    with pytest.raises(
        TypeError,
        match=f"which has type {type(param).__name__} has an unstable iteration order",
    ):
        converter_str_tupleset(param)


@pytest.mark.parametrize(
    "obj,expected",
    [
        (
            # obj
            "foo",
            # expected
            "foo",
        ),
        (
            # obj
            "foo",
            # expected
            "foo",
        ),
        (
            # obj
            b"foo",
            # expected
            "foo",
        ),
    ],
)
def test_str_ok(obj, expected):
    actual = converter_str(obj)
    assert actual == expected


@pytest.mark.parametrize(
    "obj,msg",
    [
        (
            # obj
            1,
            # msg
            "Object of type int is not a string: 1",
        ),
        (
            # obj
            ["a", "b"],
            # msg
            "Object of type list is not a string: {}".format(["a", "b"]),
        ),
        (
            # obj
            ["a"],
            # msg
            "Object of type list is not a string: {}".format(["a"]),
        ),
    ],
)
def test_str_fail(obj, msg):
    with pytest.raises(TypeError) as exc:
        converter_str(obj)
    assert str(exc.value) == msg


def test_str_rejects_none():
    with pytest.raises(TypeError) as exc:
        converter_str(None)
    assert str(exc.value) == "Object of type NoneType is not a string: None"


@pytest.mark.parametrize(
    "param,expected",
    [
        (
            # param
            "foo",
            # expected
            ("foo",),
        ),
        (
            # param
            "foo",
            # expected
            ("foo",),
        ),
        (
            # param
            b"foo",
            # expected
            (b"foo",),
        ),
        (
            # param
            ["foo", "bar"],
            # expected
            ("foo", "bar"),
        ),
        (
            # param
            ["foo", None],
            # expected
            ("foo", None),
        ),
        (
            # param
            None,
            # expected
            (),
        ),
    ],
)
def test_tuple(param, expected):
    actual = converter_tuple(param)
    assert actual == expected


@pytest.mark.parametrize(
    "s,pa_type,expected",
    [
        (
            # s
            "1",
            # pa_type
            pa.int16(),
            # expected
            1,
        ),
        (
            # s
            "1",
            # pa_type
            pa.int64(),
            # expected
            1,
        ),
        (
            # s
            "1",
            # pa_type
            pa.float32(),
            # expected
            1.0,
        ),
        (
            # s
            "1.2",
            # pa_type
            pa.float32(),
            # expected
            1.2,
        ),
        (
            # s
            "2018",
            # pa_type
            pa.timestamp("ns"),
            # expected
            pd.Timestamp("2018"),
        ),
        (
            # s
            "true",
            # pa_type
            pa.bool_(),
            # expected
            True,
        ),
        (
            # s
            "TRUE",
            # pa_type
            pa.bool_(),
            # expected
            True,
        ),
        (
            # s
            "false",
            # pa_type
            pa.bool_(),
            # expected
            False,
        ),
        (
            # s
            "t",
            # pa_type
            pa.bool_(),
            # expected
            True,
        ),
        (
            # s
            "f",
            # pa_type
            pa.bool_(),
            # expected
            False,
        ),
        (
            # s
            "1",
            # pa_type
            pa.bool_(),
            # expected
            True,
        ),
        (
            # s
            "0",
            # pa_type
            pa.bool_(),
            # expected
            False,
        ),
        (
            # s
            "yes",
            # pa_type
            pa.bool_(),
            # expected
            True,
        ),
        (
            # s
            "no",
            # pa_type
            pa.bool_(),
            # expected
            False,
        ),
        (
            # s
            "y",
            # pa_type
            pa.bool_(),
            # expected
            True,
        ),
        (
            # s
            "n",
            # pa_type
            pa.bool_(),
            # expected
            False,
        ),
        (
            # s
            " foo ",
            # pa_type
            pa.string(),
            # expected
            " foo ",
        ),
        (
            # s
            '"foo"',
            # pa_type
            pa.string(),
            # expected
            "foo",
        ),
        (
            # s
            "'foo'",
            # pa_type
            pa.string(),
            # expected
            "foo",
        ),
        (
            # s
            '"',
            # pa_type
            pa.string(),
            # expected
            '"',
        ),
        (
            # s
            '"foo',
            # pa_type
            pa.string(),
            # expected
            '"foo',
        ),
    ],
)
def test_get_str_to_python_converter_ok(s, pa_type, expected):
    converter = get_str_to_python_converter(pa_type)
    actual = converter(s)
    assert actual == expected
    assert isinstance(actual, type(expected))


def test_get_str_to_python_converter_unknown_type():
    pa_type = pa.struct([("f1", pa.int32()), ("f2", pa.bool_())])
    with pytest.raises(ValueError) as exc:
        get_str_to_python_converter(pa_type)
    assert str(exc.value) == "Cannot handle type struct<f1: int32, f2: bool>"


@pytest.mark.parametrize(
    "s,pa_type",
    [
        (
            # s
            "foo",
            # pa_type
            pa.bool_(),
        ),
        (
            # s
            "",
            # pa_type
            pa.bool_(),
        ),
        (
            # s
            "1.2",
            # pa_type
            pa.int32(),
        ),
    ],
)
def test_get_str_to_python_converter_fails(s, pa_type):
    converter = get_str_to_python_converter(pa_type)
    with pytest.raises(ValueError):
        converter(s)



import pickle
from functools import partial

import pandas as pd
import pytest

from plateau.io.dask.bag import (
    read_dataset_as_dataframe_bag,
    read_dataset_as_metapartitions_bag,
)
from plateau.io.iter import store_dataframes_as_dataset__iter
from plateau.io.testing.read import *  # noqa


@pytest.fixture(params=["dataframe"])
def output_type(request):
    return request.param


def _load_dataframes(output_type, *args, **kwargs):
    if output_type == "dataframe":
        func = read_dataset_as_dataframe_bag
    elif output_type == "metapartition":
        func = read_dataset_as_metapartitions_bag
    tasks = func(*args, **kwargs)

    s = pickle.dumps(tasks, pickle.HIGHEST_PROTOCOL)
    tasks = pickle.loads(s)

    result = tasks.compute()
    return result


@pytest.fixture()
def bound_load_dataframes(output_type):
    return partial(_load_dataframes, output_type)


def test_read_dataset_as_dataframes_partition_size(store_factory, metadata_version):
    cluster1 = pd.DataFrame(
        {"A": [1, 1], "B": [10, 10], "C": [1, 2], "Content": ["cluster1", "cluster1"]}
    )
    cluster2 = pd.DataFrame(
        {"A": [1, 1], "B": [10, 10], "C": [2, 3], "Content": ["cluster2", "cluster2"]}
    )
    cluster3 = pd.DataFrame({"A": [1], "B": [20], "C": [1], "Content": ["cluster3"]})
    cluster4 = pd.DataFrame(
        {"A": [2, 2], "B": [10, 10], "C": [1, 2], "Content": ["cluster4", "cluster4"]}
    )
    partitions = [cluster1, cluster2, cluster3, cluster4]

    store_dataframes_as_dataset__iter(
        df_generator=partitions,
        store=store_factory,
        dataset_uuid="partitioned_uuid",
        metadata_version=metadata_version,
    )
    for func in [read_dataset_as_dataframe_bag, read_dataset_as_metapartitions_bag]:
        bag = func(
            dataset_uuid="partitioned_uuid", store=store_factory, partition_size=None
        )
        assert bag.npartitions == 4
        bag = func(
            dataset_uuid="partitioned_uuid", store=store_factory, partition_size=2
        )
        assert bag.npartitions == 2



import pytest

from plateau.io.eager import garbage_collect_dataset
from plateau.io.testing.gc import *  # noqa: F403, F4


@pytest.fixture()
def garbage_collect_callable():
    return garbage_collect_dataset



# Write the benchmarking functions here.
# See "Writing benchmarks" in the asv docs for more information.


import pickle
import shutil
import sys
import tempfile
import uuid
from functools import lru_cache

import numpy as np
import pandas as pd
import pyarrow as pa
from minimalkv import get_store_from_url

from plateau.core.index import ExplicitSecondaryIndex
from plateau.io_components.metapartition import MetaPartition

from .config import AsvBenchmarkConfig


@lru_cache
def generate_partition_values(num_partitions):
    dataset_uuid = uuid.uuid1().hex
    return [
        f"{dataset_uuid}/column_1={ix % 10}/column_2={ix % 100}/{uuid.uuid1().hex}.parquet"
        for ix in range(num_partitions)
    ]


class IndexBase(AsvBenchmarkConfig):
    def setup(self, number_values, number_partitions, dtype):
        py_type, arrow_type = dtype
        self.partition_values = generate_partition_values(number_partitions)
        index_dct = {
            py_type(val): list(
                np.random.choice(self.partition_values, number_partitions // 2)
            )
            for val in range(number_values)
        }
        self.column_name = "column"
        self.ktk_index = ExplicitSecondaryIndex(
            column=self.column_name, index_dct=index_dct, dtype=arrow_type
        )
        self.tmp_dir = tempfile.mkdtemp()
        self.store = get_store_from_url(f"hfs://{self.tmp_dir}")
        self.dataset_uuid = "some_uuid"
        self.storage_key = self.ktk_index.store(self.store, self.dataset_uuid)

        self.ktk_index_not_loaded = ExplicitSecondaryIndex(
            column=self.column_name, index_storage_key=self.storage_key
        )

        self.ktk_index_loaded = self.ktk_index_not_loaded.load(self.store)

    def teardown(self, number_values, number_partitions, dtype):
        shutil.rmtree(self.tmp_dir)


class Index(IndexBase):
    params = (
        [10 * 1, 10**3],  # values
        [10 * 1, 10**3],  # partitions
        [(int, pa.int64())],  # types
    )
    param_names = ["number_values", "number_partitions", "dtype"]

    def time_remove_partitions_inplace(
        self, number_values, number_partitions, arrow_type
    ):
        partitions_to_remove = self.partition_values[len(self.partition_values) // 2 :]
        self.ktk_index.remove_partitions(partitions_to_remove, inplace=True)

    def time_load_index(self, number_values, number_partitions, arrow_type):
        self.ktk_index_not_loaded.load(self.store)

    def time_query_value(self, number_values, number_partitions, arrow_type):
        self.ktk_index.query(number_values / 2)

    def time_as_series(self, number_values, number_partitions, arrow_type):
        self.ktk_index.as_flat_series()

    def time_as_series_partitions_as_index(
        self, number_values, number_partitions, arrow_type
    ):
        self.ktk_index.as_flat_series(partitions_as_index=True)

    def time_observed_values(self, number_values, number_partitions, arrow_type):
        self.ktk_index.observed_values()


class SerializeIndex(IndexBase):
    timeout = 180
    params = (
        [(10**3, 10), (10**4, 100)],  # (values, partitions)
        [(int, pa.int64())],  # types
    )
    param_names = ["number_values__number_partitions", "dtype"]

    def setup(self, number_values__number_partitions, dtype):
        number_values, number_partitions = number_values__number_partitions
        return super().setup(number_values, number_partitions, dtype)

    def teardown(self, number_values__number_partitions, dtype):
        number_values, number_partitions = number_values__number_partitions
        return super().teardown(number_values, number_partitions, dtype)

    def track_mem_serialized(self, number_values__number_partitions, arrow_type):
        # Use `sys.getsizeof` as asv's `mem_*` just reports `0` if memory usage is low
        # enough
        return sys.getsizeof(pickle.dumps(self.ktk_index))

    def time_serialization(self, number_values__number_partitions, arrow_type):
        # Time serialization of indices
        pickle.loads(pickle.dumps(self.ktk_index))


class BuildIndex(AsvBenchmarkConfig):
    params = ([-1, 1], [10**3, 10**4], [10, 100])
    param_names = ["cardinality", "num_values", "partitions_to_merge"]

    def setup(self, cardinality, num_values, partitions_to_merge):
        self.column = "column"
        self.table = "table"
        self.merge_indices = []
        for n in range(partitions_to_merge):
            if cardinality < 0:
                array = [f"{x:010f}" for x in np.random.randn(num_values)]
            else:
                unique_vals = [f"{n:010d}" for n in range(cardinality)]
                array = [unique_vals[x % len(unique_vals)] for x in range(num_values)]
            self.df = pd.DataFrame({self.column: array})
            self.mp = MetaPartition(label=self.table, data=self.df, metadata_version=4)
            self.mp_indices = self.mp.build_indices([self.column])
            self.merge_indices.append(self.mp_indices)

    def time_metapartition_build_index(
        self, cardinality, num_values, partitions_to_merge
    ):
        self.mp.build_indices([self.column])

    def time_merge_indices(self, cardinality, num_values, partitions_to_merge):
        MetaPartition.merge_indices(self.merge_indices)



import pytest

from plateau.core.partition import Partition


def test_roundtrip():
    expected = {"files": {"Queejeb3": "file.parquet"}}
    result = Partition.from_dict("partition_label", expected).to_dict()
    assert expected == result


def test_roundtrip_no_metadata():
    expected = {"files": {"Queejeb3": "file.parquet"}}
    result = Partition.from_dict("partition_label", expected).to_dict()
    assert expected == result


def test_roundtrip_empty_metadata():
    _input = {"files": {"Queejeb3": "file.parquet"}}
    expected = {"files": {"Queejeb3": "file.parquet"}}
    result = Partition.from_dict("partition_label", _input).to_dict()
    assert expected == result


def test_raise_on_erroneous_input():
    with pytest.raises(ValueError):
        Partition.from_dict(label="label", dct="some_not_supported_external_ref")


def test_eq():
    assert not (Partition("label") == Partition("other_label"))
    assert not (Partition("label") == Partition("label", files={"some": "file"}))

    assert Partition(label="label", files={"some": "file"}) == Partition(
        label="label", files={"some": "file"}
    )



from typing import Any

PartitionDictType = dict[str, dict[str, str]]


# TODO: purge this. This is just slowing us down by creating many python objects we don't actual
# Changing the partition class needs to be done with are since it's to_dict is used for the storage metadata spec
class Partition:
    def __init__(
        self,
        label: str,
        files: dict[str, str] | None = None,
        metadata: dict | None = None,
    ):
        """An object for the internal representation of the metadata of a
        partition.

        This class is for internal use only

        Parameters
        ----------
        label:
            A label identifying the partition, e.g. `partition_1` or `P=0/L=A`
        files:
            A dictionary containing the keys of the files contained in this partition
        metadata:
            Partition level, custom metadata
        """
        self.label = label
        self.files = files if files else {}

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Partition):
            return False
        if self.label != other.label:
            return False
        if self.files != other.files:
            return False
        return True

    @staticmethod
    def from_dict(label: str, dct: str | PartitionDictType):
        if isinstance(dct, str):
            raise ValueError(
                "Trying to load a partition from a string. Probably the dataset file uses the multifile "
                "feature. Please load the metadata object using the DatasetMetadata.load_from_buffer "
                "method instead to resolve references to external partitions."
            )
        return Partition(label, files=dct.get("files", {}))

    def to_dict(self, version: Any = None) -> PartitionDictType:
        return {"files": self.files}



import pandas as pd
import pytest

from plateau.io.iter import update_dataset_from_dataframes__iter
from plateau.io.testing.update import *  # noqa


@pytest.fixture()
def bound_update_dataset():
    return _update_dataset


def _update_dataset(df_list, *args, **kwargs):
    if isinstance(df_list, pd.DataFrame):
        df_list = [df_list]
    df_generator = (x for x in df_list)
    return update_dataset_from_dataframes__iter(df_generator, *args, **kwargs)



from collections.abc import Callable, Iterable
from functools import partial
from typing import Any, cast

import pandas as pd
from minimalkv import KeyValueStore

from plateau.core.common_metadata import (
    empty_dataframe_from_schema,
    make_meta,
    store_schema_metadata,
)
from plateau.core.dataset import DatasetMetadata, DatasetMetadataBuilder
from plateau.core.docs import default_docs
from plateau.core.factory import DatasetFactory, _ensure_factory
from plateau.core.naming import (
    DEFAULT_METADATA_STORAGE_FORMAT,
    DEFAULT_METADATA_VERSION,
    PARQUET_FILE_SUFFIX,
    get_partition_file_prefix,
)
from plateau.core.typing import StoreInput
from plateau.core.utils import lazy_store
from plateau.io.iter import store_dataframes_as_dataset__iter
from plateau.io_components.delete import (
    delete_common_metadata,
    delete_indices,
    delete_top_level_metadata,
)
from plateau.io_components.gc import delete_files, dispatch_files_to_gc
from plateau.io_components.index import update_indices_from_partitions
from plateau.io_components.metapartition import (
    SINGLE_TABLE,
    MetaPartition,
    parse_input_to_metapartition,
)
from plateau.io_components.read import dispatch_metapartitions_from_factory
from plateau.io_components.update import update_dataset_from_partitions
from plateau.io_components.utils import (
    _ensure_compatible_indices,
    align_categories,
    normalize_args,
    sort_values_categorical,
    validate_partition_keys,
)
from plateau.io_components.write import raise_if_dataset_exists
from plateau.serialization import DataFrameSerializer

__all__ = (
    "delete_dataset",
    "read_dataset_as_dataframes",
    "read_table",
    "commit_dataset",
    "store_dataframes_as_dataset",
    "create_empty_dataset_header",
    "write_single_partition",
    "update_dataset_from_dataframes",
    "build_dataset_indices",
    "garbage_collect_dataset",
)


@default_docs
@normalize_args
def delete_dataset(dataset_uuid=None, store=None, factory=None):
    """Delete the entire dataset from the store.

    Parameters
    ----------
    """

    ds_factory = _ensure_factory(
        dataset_uuid=dataset_uuid,
        load_schema=False,
        store=store,
        factory=factory,
    )

    # Remove possibly unreferenced files
    garbage_collect_dataset(factory=ds_factory)

    # Delete indices first since they do not affect dataset integrity
    delete_indices(dataset_factory=ds_factory)

    for metapartition in dispatch_metapartitions_from_factory(ds_factory):
        metapartition = cast(MetaPartition, metapartition)
        metapartition.delete_from_store(dataset_uuid=dataset_uuid, store=store)

    # delete common metadata after partitions
    delete_common_metadata(dataset_factory=ds_factory)

    # Delete the top level metadata file
    delete_top_level_metadata(dataset_factory=ds_factory)


@default_docs
def read_dataset_as_dataframes(
    dataset_uuid: str | None = None,
    store=None,
    columns: dict[str, list[str]] | None = None,
    predicate_pushdown_to_io: bool = True,
    categoricals: list[str] | None = None,
    dates_as_object: bool = True,
    predicates: list[list[tuple[str, str, Any]]] | None = None,
    factory: DatasetFactory | None = None,
    dispatch_by: list[str] | None = None,
) -> list[pd.DataFrame]:
    """Read a dataset as a list of dataframes.

    Every element of the list corresponds to a physical partition.

    Parameters
    ----------

    Returns
    -------
    List[pandas.DataFrame]
        Returns a list of pandas.DataFrame. One element per partition

    Examples
    --------
    Dataset in store contains two partitions with two files each

    .. code ::

        >>> import minimalkv
        >>> from plateau.io.eager import read_dataset_as_dataframes

        >>> store = minimalkv.get_store_from_url('s3://bucket_with_dataset')

        >>> dfs = read_dataset_as_dataframes('dataset_uuid', store, 'core')
    """

    ds_factory = _ensure_factory(
        dataset_uuid=dataset_uuid,
        store=store,
        factory=factory,
    )

    mps = read_dataset_as_metapartitions(
        columns=columns,
        predicate_pushdown_to_io=predicate_pushdown_to_io,
        categoricals=categoricals,
        dates_as_object=dates_as_object,
        predicates=predicates,
        factory=ds_factory,
        dispatch_by=dispatch_by,
    )
    return [mp.data for mp in mps]


@default_docs
def read_dataset_as_metapartitions(
    dataset_uuid=None,
    store=None,
    columns=None,
    predicate_pushdown_to_io=True,
    categoricals=None,
    dates_as_object: bool = True,
    predicates=None,
    factory=None,
    dispatch_by=None,
):
    """Read a dataset as a list of
    :class:`plateau.io_components.metapartition.MetaPartition`.

    Every element of the list corresponds to a physical partition.

    Parameters
    ----------

    Returns
    -------
    List[plateau.io_components.metapartition.MetaPartition]
        Returns a tuple of the loaded dataframe and the dataset metadata

    Examples
    --------
    Dataset in store contains two partitions with two files each

    .. code ::

        >>> import minimalkv
        >>> from plateau.io.eager import read_dataset_as_dataframe

        >>> store = minimalkv.get_store_from_url('s3://bucket_with_dataset')

        >>> list_mps = read_dataset_as_metapartitions('dataset_uuid', store, 'core')
    """

    ds_factory = _ensure_factory(
        dataset_uuid=dataset_uuid,
        store=store,
        factory=factory,
    )

    from .iter import read_dataset_as_metapartitions__iterator

    ds_iter = read_dataset_as_metapartitions__iterator(
        columns=columns,
        predicate_pushdown_to_io=predicate_pushdown_to_io,
        categoricals=categoricals,
        dates_as_object=dates_as_object,
        predicates=predicates,
        factory=ds_factory,
        dispatch_by=dispatch_by,
    )
    return list(ds_iter)


@default_docs
def read_table(
    dataset_uuid: str | None = None,
    store=None,
    columns: dict[str, list[str]] | None = None,
    predicate_pushdown_to_io: bool = True,
    categoricals: list[str] | None = None,
    dates_as_object: bool = True,
    predicates: list[list[tuple[str, str, Any]]] | None = None,
    factory: DatasetFactory | None = None,
) -> pd.DataFrame:
    """A utility function to load a single table with multiple partitions as a
    single dataframe in one go. Mostly useful for smaller tables or datasets
    where all partitions fit into memory.

    The order of partitions is not guaranteed to be stable in the resulting dataframe.

    Parameters
    ----------

    Returns
    -------
    pandas.DataFrame
        Returns a pandas.DataFrame holding the data of the requested columns

    Examples
    --------
    Dataset in store contains two partitions with two files each

    .. code ::

        >>> import minimalkv
        >>> from plateau.io.eager import read_table

        >>> store = minimalkv.get_store_from_url('s3://bucket_with_dataset')

        >>> df = read_table(store, 'dataset_uuid', 'core')
    """

    ds_factory = _ensure_factory(
        dataset_uuid=dataset_uuid,
        store=store,
        factory=factory,
    )
    partitions = read_dataset_as_dataframes(
        columns=columns,
        predicate_pushdown_to_io=predicate_pushdown_to_io,
        categoricals=categoricals,
        dates_as_object=dates_as_object,
        predicates=predicates,
        factory=ds_factory,
    )

    empty_df = empty_dataframe_from_schema(
        schema=ds_factory.schema,
        columns=columns,
    )
    if categoricals:
        empty_df = empty_df.astype(dict.fromkeys(categoricals, "category"))
    dfs = list(partitions) + [empty_df]
    # require meta 4 otherwise, can't construct types/columns
    if categoricals:
        dfs = align_categories(dfs, categoricals)
    df = pd.concat(dfs, ignore_index=True, sort=False)

    # ensure column order
    if len(empty_df.columns) > 0 and list(empty_df.columns) != list(df.columns):
        df = df.reindex(empty_df.columns, copy=False, axis=1)

    return df


@default_docs
@normalize_args
def commit_dataset(
    store: StoreInput | None = None,
    dataset_uuid: str | None = None,
    new_partitions: Iterable[MetaPartition] | None = None,
    delete_scope: Iterable[dict[str, Any]] | None = None,
    metadata: dict[Any, Any] | None = None,
    metadata_merger: Callable[[list[dict]], dict] | None = None,
    default_metadata_version: int = DEFAULT_METADATA_VERSION,
    partition_on: Iterable[str] | None = None,
    factory: DatasetFactory | None = None,
    secondary_indices: Iterable[str] | None = None,
):
    """Commit new state to an existing dataset. This can be used for three
    distinct operations.

    1. Add previously written partitions to this dataset

        If for some reasons, the existing pipelines are not sufficient but you need more control, you can write the files outside of a plateau pipeline and commit them whenever you choose to.

        This should be used in combination with
        :func:`~plateau.io.eager.write_single_partition` and :func:`~plateau.io.eager.create_empty_dataset_header`.

        .. code::

            import pandas as pd
            from plateau.io.eager import write_single_partition, commit_dataset

            store = "hfs://my_store"

            # The partition writing can be done concurrently and distributed if wanted.
            # Only the information about what partitions have been written is required for the commit.
            new_partitions = [
                write_single_partition(
                    store=store,
                    dataset_uuid='dataset_uuid',
                    data=pd.DataFrame({'column': [1, 2]}),
                )
            ]

            new_dataset = commit_dataset(
                store=store,
                dataset_uuid='dataset_uuid',
                new_partitions=new_partitions,
            )

    2. Simple delete of partitions

        If you want to remove some partitions this is one of the simples ways of doing so. By simply providing a delete_scope, this removes the references to these files in an atomic commit.

        .. code::

            commit_dataset(
                store=store,
                dataset_uuid='dataset_uuid',
                delete_scope=[
                    {
                        "partition_column": "part_value_to_be_removed"
                    }
                ],
            )

    3. Add additional metadata

        To add new metadata to an existing dataset

        .. code::

            commit_dataset(
                store=store,
                dataset_uuid='dataset_uuid',
                metadata={"new": "user_metadata"},
            )

        Note::

            If you do not want the new metadata to be merged with the existing one, provide a custom ``metadata_merger``

    Parameters
    ----------
    new_partitions:
        Input partition to be committed.
    """

    if not new_partitions and not metadata and not delete_scope:
        raise ValueError(
            "Need to provide either new data, new metadata or a delete scope. None of it was provided."
        )
    if new_partitions:
        tables_in_partitions = {mp.table_name for mp in new_partitions}
        if len(tables_in_partitions) > 1:
            raise RuntimeError(
                f"Cannot commit more than one table to a dataset but got tables {sorted(tables_in_partitions)}"
            )
    if store is not None:
        store = lazy_store(store)
    ds_factory, metadata_version, partition_on = validate_partition_keys(
        dataset_uuid=dataset_uuid,
        store=store,
        ds_factory=factory,
        default_metadata_version=default_metadata_version,
        partition_on=partition_on,
    )

    mps = parse_input_to_metapartition(
        new_partitions,
        metadata_version=metadata_version,
        table_name=ds_factory.table_name,
    )

    if secondary_indices:
        mps = mps.build_indices(columns=secondary_indices)

    mps_list = [_maybe_infer_files_attribute(mp, dataset_uuid) for mp in mps]

    dmd = update_dataset_from_partitions(
        mps_list,
        store_factory=store,
        dataset_uuid=dataset_uuid,
        ds_factory=ds_factory,
        delete_scope=delete_scope,
        metadata=metadata,
        metadata_merger=metadata_merger,
    )
    return dmd


def _maybe_infer_files_attribute(metapartition, dataset_uuid):
    new_mp = metapartition.as_sentinel()
    for mp in metapartition:
        if mp.file is None:
            if mp.data is None or len(mp.data) == 0:
                raise ValueError(
                    "Trying to commit partitions without `data` or `files` information."
                    "Either one is necessary to infer the dataset tables"
                )
            new_files = {}
            for table in mp.data:
                new_files[table] = (
                    get_partition_file_prefix(
                        dataset_uuid=dataset_uuid,
                        partition_label=mp.label,
                        table=table,
                        metadata_version=mp.metadata_version,
                    )
                    + PARQUET_FILE_SUFFIX  # noqa: W503 line break before binary operator
                )
            mp = mp.copy(files=new_files)

        new_mp = new_mp.add_metapartition(mp)
    return new_mp


@default_docs
@normalize_args
def store_dataframes_as_dataset(
    store: KeyValueStore,
    dataset_uuid: str,
    dfs: list[pd.DataFrame | dict[str, pd.DataFrame]],
    metadata: dict[str, dict[str, Any]] | None = None,
    partition_on: list[str] | None = None,
    df_serializer: DataFrameSerializer | None = None,
    overwrite: bool = False,
    secondary_indices=None,
    metadata_storage_format=DEFAULT_METADATA_STORAGE_FORMAT,
    metadata_version=DEFAULT_METADATA_VERSION,
    table_name: str = SINGLE_TABLE,
):
    """Utility function to store a list of dataframes as a partitioned dataset
    with multiple tables (files).

    Useful for very small datasets where all data fits into memory.

    Parameters
    ----------
    dfs:
        The dataframe(s) to be stored.
    """
    if isinstance(dfs, pd.DataFrame):
        raise TypeError(
            f"Please pass a list of pandas.DataFrame as input. Instead got {type(dfs)}"
        )

    return store_dataframes_as_dataset__iter(
        dfs,
        store=store,
        dataset_uuid=dataset_uuid,
        metadata=metadata,
        partition_on=partition_on,
        df_serializer=df_serializer,
        overwrite=overwrite,
        secondary_indices=secondary_indices,
        metadata_storage_format=metadata_storage_format,
        metadata_version=metadata_version,
        table_name=table_name,
    )


@default_docs
@normalize_args
def create_empty_dataset_header(
    store,
    dataset_uuid,
    schema,
    partition_on=None,
    metadata=None,
    overwrite=False,
    metadata_storage_format=DEFAULT_METADATA_STORAGE_FORMAT,
    metadata_version=DEFAULT_METADATA_VERSION,
    table_name: str = SINGLE_TABLE,
):
    """Create an dataset header without any partitions. This may be used in
    combination with :func:`~plateau.io.eager.write_single_partition` to create
    implicitly partitioned datasets.

    .. note::

        The created dataset will **always** have explicit_partition==False

    .. warning::

        This function should only be used in very rare occasions. Usually you're better off using
        full end-to-end pipelines.

    Parameters
    ----------
    """
    store = lazy_store(store)()
    if not overwrite:
        raise_if_dataset_exists(dataset_uuid=dataset_uuid, store=store)

    schema = make_meta(schema, origin=table_name, partition_keys=partition_on)
    store_schema_metadata(
        schema=schema,
        dataset_uuid=dataset_uuid,
        store=store,
        table=table_name,
    )
    dataset_builder = DatasetMetadataBuilder(
        uuid=dataset_uuid,
        metadata_version=metadata_version,
        partition_keys=partition_on,
        explicit_partitions=False,
        schema=schema,
    )
    if metadata:
        for key, value in metadata.items():
            dataset_builder.add_metadata(key, value)
    if metadata_storage_format.lower() == "json":
        store.put(*dataset_builder.to_json())
    elif metadata_storage_format.lower() == "msgpack":
        store.put(*dataset_builder.to_msgpack())
    else:
        raise ValueError(
            f"Unknown metadata storage format encountered: {metadata_storage_format}"
        )
    return dataset_builder.to_dataset()


@default_docs
@normalize_args
def write_single_partition(
    store: KeyValueStore | None = None,
    dataset_uuid: str | None = None,
    data=None,
    df_serializer: DataFrameSerializer | None = None,
    metadata_version: int = DEFAULT_METADATA_VERSION,
    partition_on: list[str] | None = None,
    factory=None,
    secondary_indices=None,
    table_name: str = SINGLE_TABLE,
):
    """Write the parquet file(s) for a single partition. This will **not**
    update the dataset header and can therefore be used for highly concurrent
    dataset writes.

    For datasets with explicit partitions, the dataset header can be updated by calling
    :func:`plateau.io.eager.commit_dataset` with the output of this function.

    .. note::

        It is highly recommended to use the full pipelines whenever possible. This functionality should be
        used with caution and should only be necessary in cases where traditional pipeline scheduling is not an
        option.

    .. note::

        This function requires an existing dataset metadata file and the schemas for the tables to be present.
        Either you have ensured that the dataset always exists though some other means or use
        :func:`create_empty_dataset_header` at the start of your computation to ensure the basic dataset
        metadata is there.

    Parameters
    ----------
    data: Dict
        The input is defined according to :func:`~plateau.io_components.metapartition.parse_input_to_metapartition`

    Returns
    -------
    An empty :class:`~plateau.io_components.metapartition.MetaPartition` referencing the new files
    """
    if data is None:
        raise TypeError("The parameter `data` is not optional")
    dataset_factory, ds_metadata_version, partition_on = validate_partition_keys(
        dataset_uuid=dataset_uuid,
        store=lazy_store(store) if store else None,
        ds_factory=factory,
        default_metadata_version=metadata_version,
        partition_on=partition_on,
    )
    if dataset_factory.table_name:
        if dataset_factory.table_name != table_name:
            raise RuntimeError(
                f"Trying to write a partition with table name {table_name} but dataset {dataset_factory.dataset_uuid} has already table {dataset_factory.table_name}."
            )
    mp = parse_input_to_metapartition(
        obj=data, metadata_version=ds_metadata_version, table_name=table_name
    )

    if partition_on:
        mp = mp.partition_on(partition_on)

    if secondary_indices:
        mp = mp.build_indices(columns=secondary_indices)

    mp = mp.validate_schema_compatible(dataset_uuid=dataset_uuid, store=store)

    mp = mp.store_dataframes(
        store=store, dataset_uuid=dataset_uuid, df_serializer=df_serializer
    )
    return mp


@default_docs
@normalize_args
def update_dataset_from_dataframes(
    df_list: list[pd.DataFrame | dict[str, pd.DataFrame]],
    store: KeyValueStore | None = None,
    dataset_uuid: str | None = None,
    delete_scope=None,
    metadata=None,
    df_serializer: DataFrameSerializer | None = None,
    metadata_merger: Callable | None = None,
    default_metadata_version: int = DEFAULT_METADATA_VERSION,
    partition_on: list[str] | None = None,
    sort_partitions_by: str | None = None,
    secondary_indices: list[str] | None = None,
    table_name: str = SINGLE_TABLE,
    factory: DatasetFactory | None = None,
) -> DatasetMetadata:
    """Update a plateau dataset in store at once, using a list of dataframes.

    Useful for datasets which do not fit into memory.

    Parameters
    ----------
    df_list:
        The dataframe(s) to be stored.

    Returns
    -------
    The dataset metadata object (:class:`~plateau.core.dataset.DatasetMetadata`).

    See Also
    --------
    :ref:`mutating_datasets`
    """
    ds_factory, metadata_version, partition_on = validate_partition_keys(
        dataset_uuid=dataset_uuid,
        store=store,
        ds_factory=factory,
        default_metadata_version=default_metadata_version,
        partition_on=partition_on,
    )

    # ensured by normalize_args but mypy doesn't recognize it
    assert secondary_indices is not None

    inferred_indices = _ensure_compatible_indices(ds_factory, secondary_indices)
    del secondary_indices

    mp = parse_input_to_metapartition(
        df_list,
        metadata_version=metadata_version,
        table_name=table_name,
    )

    if sort_partitions_by:
        mp = mp.apply(partial(sort_values_categorical, columns=sort_partitions_by))

    if partition_on:
        mp = mp.partition_on(partition_on)

    if inferred_indices:
        mp = mp.build_indices(inferred_indices)

    mp = mp.store_dataframes(
        store=store, dataset_uuid=dataset_uuid, df_serializer=df_serializer
    )

    return update_dataset_from_partitions(
        mp,
        store_factory=store,
        dataset_uuid=dataset_uuid,
        ds_factory=ds_factory,
        delete_scope=delete_scope,
        metadata=metadata,
        metadata_merger=metadata_merger,
    )


@default_docs
@normalize_args
def build_dataset_indices(store, dataset_uuid, columns, factory=None):
    """Function which builds a
    :class:`~plateau.core.index.ExplicitSecondaryIndex`.

    This function loads the dataset, computes the requested indices and writes
    the indices to the dataset. The dataset partitions itself are not mutated.

    Parameters
    ----------
    """
    ds_factory = _ensure_factory(
        dataset_uuid=dataset_uuid,
        store=store,
        factory=factory,
    )

    if ds_factory.schema is None:
        raise ValueError("Dataset schema needs to be loaded for column selection.")
    cols_to_load = set(columns) & set(ds_factory.schema.names)

    new_partitions = []
    for mp in dispatch_metapartitions_from_factory(ds_factory):
        mp = mp.load_dataframes(
            store=ds_factory.store,
            columns=cols_to_load,
        )
        mp = mp.build_indices(columns=columns)
        mp = mp.remove_dataframes()  # Remove dataframe from memory
        new_partitions.append(mp)

    return update_indices_from_partitions(
        new_partitions, dataset_metadata_factory=ds_factory
    )


@default_docs
@normalize_args
def garbage_collect_dataset(dataset_uuid=None, store=None, factory=None):
    """Remove auxiliary files that are no longer tracked by the dataset.

    These files include indices that are no longer referenced by the metadata
    as well as files in the directories of the tables that are no longer
    referenced. The latter is only applied to static datasets.

    Parameters
    ----------
    """

    ds_factory = _ensure_factory(
        dataset_uuid=dataset_uuid,
        store=store,
        factory=factory,
    )

    nested_files = dispatch_files_to_gc(
        dataset_uuid=None, store_factory=None, chunk_size=None, factory=ds_factory
    )

    # Given that `nested_files` is a generator with a single element, just
    # return the output of `delete_files` on that element.
    return delete_files(next(nested_files), store_factory=ds_factory.store_factory)



import pickle

import dask
import pytest

from plateau.io.dask.delayed import garbage_collect_dataset__delayed
from plateau.io.testing.gc import *  # noqa


def _run_garbage_collect(*args, **kwargs):
    tasks = garbage_collect_dataset__delayed(*args, **kwargs)
    s = pickle.dumps(tasks, pickle.HIGHEST_PROTOCOL)
    tasks = pickle.loads(s)
    dask.compute(tasks)


@pytest.fixture()
def garbage_collect_callable():
    return _run_garbage_collect



import tempfile
import uuid

from minimalkv import get_store_from_url

from plateau.core.common_metadata import make_meta
from plateau.core.testing import get_dataframe_alltypes
from plateau.io_components.metapartition import MetaPartition
from plateau.io_components.write import (
    persist_common_metadata,
    store_dataset_from_partitions,
)

from .config import AsvBenchmarkConfig


def generate_mp():
    return MetaPartition(
        label=uuid.uuid4().hex,
        schema=make_meta(get_dataframe_alltypes(), origin="alltypes"),
        file="fakefile",
    )


def _partition_metadata(depth, num_leaves):
    if depth == 0:
        return "the_end"
    part_meta = _partition_metadata(depth - 1, num_leaves)
    return {str(k): part_meta for k in range(num_leaves)}


def generate_metadata(max_depth=7, num_leaves=5):
    """Generate a metadata dictionary which holds many `partition_metadata`
    keys."""
    return {
        "creation_time": "2018-05-05 12:00:00",
        "partition_metadata": _partition_metadata(max_depth, num_leaves),
    }


class TimeStoreDataset(AsvBenchmarkConfig):
    timeout = 120
    params = ([10, 10**2, 10**3], [4], [2, 4])
    param_names = ["num_partitions", "max_depth", "num_leaves"]

    def setup(self, num_partitions, max_depth, num_leaves):
        self.store = get_store_from_url(f"hfs://{tempfile.mkdtemp()}")
        self.partitions = [generate_mp() for _ in range(num_partitions)]
        self.dataset_uuid = "dataset_uuid"
        self.user_dataset_metadata = {}

    def time_store_dataset_from_partitions(self, num_partitions, max_depth, num_leaves):
        store_dataset_from_partitions(
            partition_list=self.partitions,
            store=self.store,
            dataset_uuid=self.dataset_uuid,
            dataset_metadata=self.user_dataset_metadata,
        )


class TimePersistMetadata(AsvBenchmarkConfig):
    timeout = 240
    params = [1, 10**2, 10**3]

    def setup(self, num_partitions):
        self.store = get_store_from_url(f"hfs://{tempfile.mkdtemp()}")
        self.schemas = [generate_mp().schema for _ in range(num_partitions)]
        self.dataset_uuid = "dataset_uuid"

    def time_persist_common_metadata(self, num_partitions):
        persist_common_metadata(
            self.schemas, None, self.store, self.dataset_uuid, "name"
        )



import logging

from plateau.core import naming
from plateau.core.index import ExplicitSecondaryIndex
from plateau.io_components.metapartition import MetaPartition
from plateau.io_components.write import persist_indices

_logger = logging.getLogger(__name__)


def update_indices_from_partitions(partition_list, dataset_metadata_factory):
    """This takes indices from a partition list and overwrites all indices in
    the dataset metadata provided by the dataset metadata factory.

    The same is done in the store dataset part. This is used in an
    additional build index step (by the build_dataset_indices__pipeline)
    which should be used after updating partitions of a dataset.
    """

    dataset_indices = MetaPartition.merge_indices(partition_list)

    indices = persist_indices(
        store=dataset_metadata_factory.store,
        dataset_uuid=dataset_metadata_factory.uuid,
        indices=dataset_indices,
    )

    for column, storage_key in indices.items():
        dataset_metadata_factory.indices[column] = ExplicitSecondaryIndex(
            column=column, index_storage_key=storage_key
        )

    dataset_metadata_factory.store.put(
        naming.metadata_key_from_uuid(dataset_metadata_factory.uuid),
        dataset_metadata_factory.to_json(),
    )
    return dataset_metadata_factory



#!/usr/bin/env python
"""This module contains functionality for persisting/serialising DataFrames."""

import gzip
from collections.abc import Iterable
from io import BufferedIOBase, BytesIO, StringIO
from typing import Any

import pandas as pd
import pyarrow as pa
from minimalkv import KeyValueStore
from pandas.errors import EmptyDataError

from ._generic import (
    DataFrameSerializer,
    PredicatesType,
    check_predicates,
    filter_df,
    filter_df_from_predicates,
)


class CsvSerializer(DataFrameSerializer):
    def __init__(self, compress=True):
        self.compress = compress

    def __eq__(self, other):
        return isinstance(other, CsvSerializer) and (self.compress == other.compress)

    def __repr__(self):
        return f"CsvSerializer(compress={self.compress!r})"

    @staticmethod
    def restore_dataframe(
        store: KeyValueStore,
        key: str,
        filter_query: str | None = None,
        columns: Iterable[str] | None = None,
        predicate_pushdown_to_io: Any = None,
        categories: Iterable[str] | None = None,
        predicates: PredicatesType | None = None,
        date_as_object: Any = None,
        **kwargs,
    ):
        check_predicates(predicates)
        compression: str | None
        if key.endswith(".csv.gz"):
            compression = "gzip"
        elif key.endswith(".csv"):
            compression = None

        if (not columns) and (columns is not None):
            # pd.read_csv does not seem to support proper reads w/o columns (it returns a DF w/o any row)
            columns = None
            project_to_no_cols = True
        else:
            project_to_no_cols = False

        dtype: dict[str, str] | None
        if categories:
            dtype = dict.fromkeys(categories, "category")
        else:
            dtype = None

        try:
            df = pd.read_csv(
                BytesIO(store.get(key)),
                compression=compression,
                sep=";",
                encoding="utf-8",
                usecols=columns,
                dtype=dtype,
            )
            if project_to_no_cols:
                df = df[[]]
            if len(df) == 0:
                # in that case, Pandas decided to use a weird index type, let's fix that
                df.index = pd.RangeIndex(start=0, stop=0, step=1)
        except EmptyDataError:
            df = pd.DataFrame()

        if predicates:
            return filter_df_from_predicates(df, predicates)
        else:
            return filter_df(df, filter_query)

    def store(self, store, key_prefix, df):
        if isinstance(df, pa.Table):
            # Prior to pyarrow 13.0.0 coerce_temporal_nanoseconds didn't exist
            # as it was introduced for backwards compatibility with pandas 1.x
            _coerce = {"coerce_temporal_nanoseconds": True}
            df = df.to_pandas(**_coerce)

        key = f"{key_prefix}.csv"
        result_stream = BytesIO()
        iostream: BufferedIOBase
        if self.compress:
            iostream = gzip.GzipFile(fileobj=result_stream, mode="wb")
            key += ".gz"
        else:
            iostream = result_stream

        # It seems that df.to_csv always writes default str, and thus
        # needs a StringIO in and a BytesIO in Python 2.
        # if someone finds a neat solution to do this in a compatible way,
        # please go ahead:

        unicode_stream = StringIO()
        if "0.23" <= pd.__version__ < "0.24":
            unicode_stream.write(df.to_csv(index=False, sep=";"))
        else:
            df.to_csv(unicode_stream, index=False, sep=";")
        iostream.write(unicode_stream.getvalue().encode("utf-8"))

        if self.compress:
            iostream.close()
        store.put(key, result_stream.getvalue())
        return key



import logging
from functools import partial

import dask
import dask.dataframe as dd
import pandas as pd

from plateau.core._compat import PANDAS_3

_logger = logging.getLogger()
_PAYLOAD_COL = "__ktk_shuffle_payload"

try:
    # Technically distributed is an optional dependency
    from distributed.protocol import deserialize_bytes, serialize_bytes

    HAS_DISTRIBUTED = True
except ImportError:
    HAS_DISTRIBUTED = False
    serialize_bytes = None
    deserialize_bytes = None

__all__ = (
    "pack_payload_pandas",
    "pack_payload",
    "unpack_payload_pandas",
    "unpack_payload",
)


def pack_payload_pandas(partition: pd.DataFrame, group_key: list[str]) -> pd.DataFrame:
    if not HAS_DISTRIBUTED:
        _logger.warning(
            "Shuffle payload columns cannot be compressed since distributed is not installed."
        )
        return partition

    if partition.empty:
        res = partition[group_key]
        res[_PAYLOAD_COL] = b""
    else:
        res = partition.groupby(
            group_key,
            sort=False,
            observed=True,
            # Keep the as_index s.t. the group values are not dropped. With this
            # the behaviour seems to be consistent along pandas versions
            as_index=True,
        ).apply(lambda x: pd.Series({_PAYLOAD_COL: serialize_bytes(x)}))

        res = res.reset_index()
    return res


def pack_payload(df: dd.DataFrame, group_key: list[str] | str) -> dd.DataFrame:
    """Pack all payload columns (everything except of group_key) into a single
    columns. This column will contain a single byte string containing the
    serialized and compressed payload data. The payload data is just dead
    weight when reshuffling. By compressing it once before the shuffle starts,
    this saves a lot of memory and network/disk IO.

    Example::

        >>> import pandas as pd
        ... import dask.dataframe as dd
        ... from dask.dataframe.shuffle import pack_payload
        ...
        ... df = pd.DataFrame({"A": [1, 1] * 2 + [2, 2] * 2 + [3, 3] * 2, "B": range(12)})
        ... ddf = dd.from_pandas(df, npartitions=2)

        >>> ddf.partitions[0].compute()

        A  B
        0  1  0
        1  1  1
        2  1  2
        3  1  3
        4  2  4
        5  2  5

        >>> pack_payload(ddf, "A").partitions[0].compute()

        A                               __dask_payload_bytes
        0  1  b'\x03\x00\x00\x00\x00\x00\x00\x00)\x00\x00\x03...
        1  2  b'\x03\x00\x00\x00\x00\x00\x00\x00)\x00\x00\x03...


    See also https://github.com/dask/dask/pull/6259
    """

    if (
        # https://github.com/pandas-dev/pandas/issues/34455
        df._meta.index.dtype == "float64"
        # TODO: Try to find out what's going on an file a bug report
        # For datetime indices the apply seems to be corrupt
        # s.t. apply(lambda x:x) returns different values
        or (df._meta.index.dtype == "datetime64[ns]")
    ):
        return df
    if not HAS_DISTRIBUTED:
        _logger.warning(
            "Shuffle payload columns cannot be compressed since distributed is not installed."
        )
        return df

    if not isinstance(group_key, list):
        group_key = [group_key]

    packed_meta = df._meta[group_key]
    packed_meta[_PAYLOAD_COL] = b""

    _pack_payload = partial(pack_payload_pandas, group_key=group_key)

    with dask.config.set({"dataframe.convert-string": False}):
        return df.map_partitions(_pack_payload, meta=packed_meta)


def unpack_payload_pandas(
    partition: pd.DataFrame, unpack_meta: pd.DataFrame
) -> pd.DataFrame:
    """Revert ``pack_payload_pandas`` and restore packed payload.

    unpack_meta:
        A dataframe indicating the schema of the unpacked data. This will be returned in case the input is empty
    """
    if not HAS_DISTRIBUTED:
        _logger.warning(
            "Shuffle payload columns cannot be compressed since distributed is not installed."
        )
        return partition

    if partition.empty:
        return unpack_meta.iloc[:0]
    group_cols = list(set(partition.columns) - {_PAYLOAD_COL})

    def _inner(partition):
        return pd.concat(
            partition[_PAYLOAD_COL].map(deserialize_bytes).values,
            ignore_index=True,
        )

    if not group_cols:
        return _inner(partition)
    return (
        partition.groupby(
            group_cols,
            sort=False,
            observed=True,
        )
        .apply(_inner)
        .reset_index(drop=not PANDAS_3)[unpack_meta.columns]
    )


def unpack_payload(df: dd.DataFrame, unpack_meta: pd.DataFrame) -> dd.DataFrame:
    """Revert payload packing of ``pack_payload`` and restores full
    dataframe."""

    if (
        # https://github.com/pandas-dev/pandas/issues/34455
        (df._meta.index.dtype == "float64")
        # TODO: Try to find out what's going on an file a bug report
        # For datetime indices the apply seems to be corrupt
        # s.t. apply(lambda x:x) returns different values
        or df._meta.index.dtype == "datetime64[ns]"
    ):
        return df

    if not HAS_DISTRIBUTED:
        _logger.warning(
            "Shuffle payload columns cannot be compressed since distributed is not installed."
        )
        return df

    with dask.config.set({"dataframe.convert-string": False}):
        return df.map_partitions(
            unpack_payload_pandas, unpack_meta=unpack_meta, meta=unpack_meta
        )



"""This module is a collection of tests which should be implemented by all
plateau.

**read** backends. The tests are not subject to the semantic versioning scheme and may change with minor or even patch releases.

To use the tests of this module, add the following import statement to your test module and ensure that the following fixtures are available in your test environment.

```
from plateau.io.testing.read import *  # noqa
```

Fixtures required to be implemented:

* ``output_type`` - One of {`dataframe`, `metpartition`, `table`} to define the outptu type of the returned result.
* ``bound_load_dataframes`` - A callable which will retrieve the partitions in the format specified by ``output_type``. The callable should accept all keyword arguments expected for a plateau reader.

Source test data

* ``dataset`` - A fixture generating test data (TODO: Expose this as a testing function)
* ``store_factory`` - A function scoped store factory
* ``store_session_factory`` - A session scoped store factory

Feature toggles (optional):

The following fixtures should be present (see tests.read.conftest)
* ``use_categoricals`` - Whether or not the call retrievs categorical data.
* ``dates_as_object`` - Whether or not the call retrievs date columns as objects.
"""

import datetime
from functools import partial
from itertools import permutations

import pandas as pd
import pandas.testing as pdt
import pytest
from minimalkv import get_store_from_url

from plateau.core._compat import PANDAS_3, pandas_infer_string
from plateau.io.eager import store_dataframes_as_dataset
from plateau.io.iter import store_dataframes_as_dataset__iter
from plateau.io_components.metapartition import SINGLE_TABLE, MetaPartition


@pytest.fixture(params=[True, False], ids=["use_categoricals", "no_categoricals"])
def use_categoricals(request):
    return request.param


@pytest.fixture(params=[True, False], ids=["dates_as_object", "datest_as_datetime"])
def dates_as_object(request):
    return request.param


@pytest.fixture(params=[True, False], ids=["use_factory", "no_factory"])
def use_dataset_factory(request, dates_as_object):
    return request.param


class NoPickle:
    def __getstate__(self):
        raise RuntimeError("do NOT pickle this object!")


def mark_nopickle(obj):
    obj._nopickle = NoPickle()


def no_pickle_store(url):
    store = get_store_from_url(url)
    mark_nopickle(store)
    return store


def no_pickle_factory(url):
    return partial(no_pickle_store, url)


@pytest.fixture(params=["URL", "KeyValue", "Callable"])
def store_input_types(request, tmpdir):
    url = f"hfs://{tmpdir}"

    if request.param == "URL":
        return url
    elif request.param == "KeyValue":
        return get_store_from_url(url)
    elif request.param == "Callable":
        return no_pickle_factory(url)
    else:
        raise RuntimeError(f"Encountered unknown store type {type(request.param)}")


def test_store_input_types(store_input_types, bound_load_dataframes):
    from plateau.io.eager import store_dataframes_as_dataset
    from plateau.serialization.testing import get_dataframe_not_nested

    dataset_uuid = "dataset_uuid"
    df = get_dataframe_not_nested(10)

    store_dataframes_as_dataset(
        dfs=[df],
        dataset_uuid=dataset_uuid,
        store=store_input_types,
        partition_on=[df.columns[0]],
        secondary_indices=[df.columns[1]],
    )

    # Use predicates to trigger partition pruning with indices
    predicates = [
        [
            (df.columns[0], "==", df.loc[0, df.columns[0]]),
            (df.columns[1], "==", df.loc[0, df.columns[1]]),
        ]
    ]

    result = bound_load_dataframes(
        dataset_uuid=dataset_uuid,
        store=store_input_types,
        predicates=predicates,
        dates_as_object=True,
    )

    if isinstance(result, list):
        result = result[0]

    if isinstance(result, MetaPartition):
        result = result.data

    if isinstance(result, dict):
        result = result[SINGLE_TABLE]

    pdt.assert_frame_equal(result, df.head(1), check_dtype=False)


def _perform_read_test(
    dataset_uuid,
    store_factory,
    execute_read_callable,
    use_categoricals,
    output_type,
    dates_as_object,
    read_kwargs=None,
    ds_factory=None,
):
    if not read_kwargs:
        read_kwargs = {}
    if use_categoricals:
        # dataset_with_index has an index on L but not on P
        categoricals = ["P", "L"]
    else:
        categoricals = None

    result = execute_read_callable(
        dataset_uuid=dataset_uuid,
        store=store_factory,
        factory=ds_factory,
        categoricals=categoricals,
        dates_as_object=dates_as_object,
        **read_kwargs,
    )

    assert len(result) == 2

    if output_type == "metapartition":
        for res in result:
            assert isinstance(res, MetaPartition)
        result = [mp.data for mp in result]

        def sort_by(obj):
            return obj[SINGLE_TABLE].P.iloc[0]

    elif output_type == "table":
        assert isinstance(result[0], pd.DataFrame)
        assert "P" in result[0]

        def sort_by(obj):
            return obj.P.iloc[0]

    else:
        assert isinstance(result[0], pd.DataFrame)
        assert "P" in result[0]

        def sort_by(obj):
            return obj.P.iloc[0]

    result = sorted(result, key=sort_by)

    expected_df_core_1 = pd.DataFrame(
        {"P": [1], "L": [1], "TARGET": [1], "DATE": [datetime.date(2010, 1, 1)]}
    )
    expected_df_core_2 = pd.DataFrame(
        {"P": [2], "L": [2], "TARGET": [2], "DATE": [datetime.date(2009, 12, 31)]}
    )
    expected_dfs = [
        expected_df_core_1,
        expected_df_core_2,
    ]

    for res, expected_df_core in zip(result, expected_dfs, strict=False):
        if not dates_as_object:
            expected_df_core["DATE"] = pd.to_datetime(expected_df_core["DATE"])
        if use_categoricals:
            expected_df_core = expected_df_core.astype(
                {"P": "category", "L": "category"}
            )

        pdt.assert_frame_equal(
            res.reset_index(drop=True),
            expected_df_core.reset_index(drop=True),
            check_dtype=False,
            check_like=True,
            check_categorical=False,
        )


@pytest.mark.parametrize(
    "predicates",
    [
        [[("P", "==", 2)]],
        [[("P", "in", [2])]],
        [[("P", "!=", 1)]],
        [[("P", ">", 1)]],
        [[("P", ">=", 2)]],
    ],
)
def test_read_dataset_as_dataframes_predicate(
    dataset, store_session_factory, bound_load_dataframes, predicates, output_type
):
    if output_type != "dataframe":
        pytest.skip()
    result = bound_load_dataframes(
        dataset_uuid=dataset.uuid,
        store=store_session_factory,
        predicates=predicates,
    )
    core_result = pd.concat(result)

    expected_core = pd.DataFrame(
        {"P": [2], "L": [2], "TARGET": [2], "DATE": [datetime.date(2009, 12, 31)]}
    )
    pdt.assert_frame_equal(
        core_result, expected_core, check_dtype=False, check_like=True
    )


@pytest.mark.parametrize(
    "predicates",
    [
        [[("P", "==", 2), ("TARGET", "==", 2)]],
        [[("P", "in", [2]), ("TARGET", "==", 2)]],
        [[("P", "!=", 1), ("L", "==", 2)]],
        [[("P", "!=", 1), ("L", "in", [2])]],
        [[("P", ">", 2)], [("TARGET", ">=", 2)]],
        [[("P", ">=", 2)], [("TARGET", ">=", 2)]],
    ],
)
def test_read_dataset_as_dataframes_predicate_with_partition_keys(
    dataset_partition_keys,
    store_session_factory,
    bound_load_dataframes,
    predicates,
    output_type,
):
    if output_type != "dataframe":
        pytest.skip()
    result = bound_load_dataframes(
        dataset_uuid=dataset_partition_keys.uuid,
        store=store_session_factory,
        predicates=predicates,
    )

    core_result = pd.concat(result)

    expected_core = pd.DataFrame(
        {"P": [2], "L": [2], "TARGET": [2], "DATE": [datetime.date(2009, 12, 31)]}
    )
    pdt.assert_frame_equal(
        core_result, expected_core, check_dtype=False, check_like=True
    )


def test_read_dataset_as_dataframes_predicate_empty(
    dataset_partition_keys,
    store_session_factory,
    output_type,
    bound_load_dataframes,
):
    if output_type != "dataframe":
        pytest.skip()
    result = bound_load_dataframes(
        dataset_uuid=dataset_partition_keys.uuid,
        store=store_session_factory,
        predicates=[[("P", "==", -42)]],
        columns={SINGLE_TABLE: ["P", "L", "TARGET"]},
    )
    assert len(result) == 0


def _gen_partition(b_c):
    b, c = b_c
    return pd.DataFrame({"a": [1], "b": [b], "c": c})


def test_read_dataset_as_dataframes_dispatch_by_empty(
    store_session_factory,
    dataset_dispatch_by,
    bound_load_dataframes,
    backend_identifier,
    output_type,
    metadata_version,
    dataset_dispatch_by_uuid,
):
    if output_type == "table":
        pytest.skip()
    # Dispatch by primary index "A"
    dispatched = bound_load_dataframes(
        dataset_uuid=dataset_dispatch_by_uuid,
        store=store_session_factory,
        dispatch_by=[],
    )

    assert len(dispatched) == 1


@pytest.mark.parametrize("dispatch_by", ["A", "B", "C"])
def test_read_dataset_as_dataframes_dispatch_by_single_col(
    store_session_factory,
    dataset_dispatch_by,
    bound_load_dataframes,
    backend_identifier,
    dispatch_by,
    output_type,
    metadata_version,
    dataset_dispatch_by_uuid,
):
    if output_type == "table":
        pytest.skip()
    # Dispatch by primary index "A"
    dispatched_a = bound_load_dataframes(
        dataset_uuid=dataset_dispatch_by_uuid,
        store=store_session_factory,
        dispatch_by=[dispatch_by],
    )

    unique_a = set()
    for data in dispatched_a:
        unique_dispatch = data[dispatch_by].unique()
        assert len(unique_dispatch) == 1
        assert unique_dispatch[0] not in unique_a
        unique_a.add(unique_dispatch[0])


@pytest.fixture(scope="session")
def dataset_dispatch_by_uuid():
    import uuid

    return uuid.uuid1().hex


@pytest.fixture(scope="session")
def dataset_dispatch_by(
    metadata_version, store_session_factory, dataset_dispatch_by_uuid
):
    cluster1 = pd.DataFrame(
        {"A": [1, 1], "B": [10, 10], "C": [1, 2], "Content": ["cluster1", "cluster1"]}
    )
    cluster2 = pd.DataFrame(
        {"A": [1, 1], "B": [10, 10], "C": [2, 3], "Content": ["cluster2", "cluster2"]}
    )
    cluster3 = pd.DataFrame({"A": [1], "B": [20], "C": [1], "Content": ["cluster3"]})
    cluster4 = pd.DataFrame(
        {"A": [2, 2], "B": [10, 10], "C": [1, 2], "Content": ["cluster4", "cluster4"]}
    )
    clusters = [cluster1, cluster2, cluster3, cluster4]

    store_dataframes_as_dataset__iter(
        df_generator=clusters,
        store=store_session_factory,
        dataset_uuid=dataset_dispatch_by_uuid,
        metadata_version=metadata_version,
        partition_on=["A", "B"],
        secondary_indices=["C"],
    )
    return pd.concat(clusters).sort_values(["A", "B", "C"]).reset_index(drop=True)


def test_read_dataset_as_dataframes_dispatch_by_multi_col(
    store_session_factory,
    bound_load_dataframes,
    output_type,
    dataset_dispatch_by,
    dataset_dispatch_by_uuid,
):
    if output_type == "table":
        pytest.skip()
    for dispatch_by in permutations(("A", "B", "C"), 2):
        dispatched = bound_load_dataframes(
            dataset_uuid=dataset_dispatch_by_uuid,
            store=store_session_factory,
            dispatch_by=dispatch_by,
        )
        uniques = pd.DataFrame(columns=dispatch_by)
        for part in dispatched:
            if isinstance(part, MetaPartition):
                data = part.data
            else:
                data = part
            unique_dispatch = data[list(dispatch_by)].drop_duplicates()
            assert len(unique_dispatch) == 1
            uniques = pd.concat([uniques, unique_dispatch])
        assert not any(uniques.duplicated())


@pytest.mark.parametrize(
    "dispatch_by, predicates, expected_dispatches",
    [
        # This should only dispatch one partition since there is only
        # one file with valid data points
        (["A"], [[("C", ">", 2)]], 1),
        # We dispatch and restrict to one valie, i.e. one dispatch
        (["B"], [[("B", "==", 10)]], 1),
        # The same is true for a non-partition index col
        (["C"], [[("C", "==", 1)]], 1),
        # A condition where both primary and secondary indices need to work together
        (["A", "C"], [[("A", ">", 1), ("C", "<", 3)]], 2),
    ],
)
def test_read_dispatch_by_with_predicates(
    store_session_factory,
    dataset_dispatch_by_uuid,
    bound_load_dataframes,
    dataset_dispatch_by,
    dispatch_by,
    output_type,
    expected_dispatches,
    predicates,
):
    if output_type == "table":
        pytest.skip()

    dispatched = bound_load_dataframes(
        dataset_uuid=dataset_dispatch_by_uuid,
        store=store_session_factory,
        dispatch_by=dispatch_by,
        predicates=predicates,
    )

    assert len(dispatched) == expected_dispatches, dispatched


def test_read_dataset_as_dataframes(
    dataset,
    store_session_factory,
    dataset_factory,
    use_dataset_factory,
    bound_load_dataframes,
    use_categoricals,
    output_type,
    dates_as_object,
):
    if use_dataset_factory:
        dataset_uuid = dataset.uuid
        store_factory = store_session_factory
        ds_factory = None
    else:
        dataset_uuid = None
        store_factory = None
        ds_factory = dataset_factory

    _perform_read_test(
        dataset_uuid=dataset_uuid,
        store_factory=store_factory,
        ds_factory=ds_factory,
        execute_read_callable=bound_load_dataframes,
        use_categoricals=use_categoricals,
        output_type=output_type,
        dates_as_object=dates_as_object,
    )


def test_read_dataset_as_dataframes_columns_projection(
    store_factory, bound_load_dataframes, metadata_version
):
    def _f(b_c):
        b, c = b_c
        df = pd.DataFrame({"a": [1, 1], "b": [b, b], "c": c, "d": [b, b + 1]})
        return df

    in_partitions = [_f([1, 100])]
    dataset_uuid = "partitioned_uuid"
    store_dataframes_as_dataset(
        dfs=in_partitions,
        store=store_factory,
        dataset_uuid=dataset_uuid,
        metadata_version=metadata_version,
        partition_on=["a", "b"],
    )

    result = bound_load_dataframes(
        dataset_uuid=dataset_uuid,
        store=store_factory,
        columns=["a", "b", "c"],
    )
    probe = result[0]

    if isinstance(probe, MetaPartition):
        result_dfs = [mp.data for mp in result]
    else:
        result_dfs = result
    result_df = pd.concat(result_dfs).reset_index(drop=True)

    expected_df = pd.DataFrame({"a": [1, 1], "b": [1, 1], "c": [100, 100]})
    pdt.assert_frame_equal(expected_df, result_df, check_like=True)


def test_read_dataset_as_dataframes_columns_primary_index_only(
    store_factory, bound_load_dataframes, metadata_version
):
    def _f(b_c):
        b, c = b_c
        df = pd.DataFrame({"a": [1, 1], "b": [b, b], "c": c, "d": [b, b + 1]})
        return df

    in_partitions = [_f([1, 100])]
    dataset_uuid = "partitioned_uuid"

    store_dataframes_as_dataset(
        dfs=in_partitions,
        store=store_factory,
        dataset_uuid=dataset_uuid,
        metadata_version=metadata_version,
        partition_on=["a", "b"],
    )
    result = bound_load_dataframes(
        dataset_uuid=dataset_uuid, store=store_factory, columns=["a", "b"]
    )
    probe = result[0]

    if isinstance(probe, MetaPartition):
        result_dfs = [mp.data for mp in result]
    else:
        result_dfs = result
    result_df = pd.concat(result_dfs).reset_index(drop=True)

    expected_df = pd.DataFrame({"a": [1, 1], "b": [1, 1]})
    pdt.assert_frame_equal(expected_df, result_df, check_like=True)


def test_empty_predicate_pushdown_empty_col_projection(
    dataset, store_session_factory, bound_load_dataframes, backend_identifier
):
    result = bound_load_dataframes(
        dataset_uuid=dataset.uuid,
        store=store_session_factory,
        columns=[],
        predicates=[[("P", "==", 12345678)]],  # this product doesn't exist
    )

    if backend_identifier.startswith("dask"):
        pytest.xfail("Output of dask for empty results is currently inconsistent")
    probe = result[0]

    if isinstance(probe, MetaPartition):
        result_dfs = [mp.data for mp in result]
    else:
        result_dfs = result
    res = pd.concat(result_dfs).reset_index(drop=True)
    assert len(res) == 0

    # Ignore the different types of indices pandas generates after `reset_index`.
    pdt.assert_frame_equal(
        res,
        pd.DataFrame(
            index=res.index,
            columns=pd.Index([], dtype=str if pandas_infer_string() else "object"),
        ),
    )


@pytest.mark.parametrize("partition_on", [["a", "b"], ["c"], ["a", "b", "c"]])
@pytest.mark.parametrize("datetype", [datetime.datetime, datetime.date])
@pytest.mark.parametrize("comp", ["==", ">="])
def test_datetime_predicate_with_dates_as_object(
    dataset,
    store_factory,
    bound_load_dataframes,
    metadata_version,
    output_type,
    partition_on,
    datetype,
    comp,
):
    def _f(b_c):
        b, c = b_c
        df = pd.DataFrame({"a": [1, 1], "b": [b, b], "c": c, "d": [b, b + 1]})

        # Account for pandas 2.1 change in behaviour in which datetime.datetime
        # columns have their units set to [us] rather than [ns].
        return (
            df.astype({"c": "datetime64[ns]"})
            if df["c"].dtype == "datetime64[us]"
            else df
        )

    in_partitions = [_f([1, datetype(2000, 1, 1)])]
    dataset_uuid = "partitioned_uuid"
    store_dataframes_as_dataset(
        dfs=in_partitions,
        store=store_factory,
        dataset_uuid=dataset_uuid,
        metadata_version=metadata_version,
        partition_on=partition_on,
    )

    result = bound_load_dataframes(
        dataset_uuid="partitioned_uuid",
        store=store_factory,
        predicates=[[("c", comp, datetype(2000, 1, 1))]],
        dates_as_object=True,
    )
    if output_type != "dataframe":
        return

    assert len(result) == 1
    df_actual = result[0]

    df_expected = in_partitions[0]
    pdt.assert_frame_equal(df_actual, df_expected, check_like=True)


def test_binary_column_metadata(store_factory, bound_load_dataframes):
    df = pd.DataFrame({b"int_col": [1], "🙈".encode(): [2]})

    store_dataframes_as_dataset(
        dfs=[df], store=store_factory, dataset_uuid="dataset_uuid"
    )

    result = bound_load_dataframes(dataset_uuid="dataset_uuid", store=store_factory)

    probe = result[0]
    if isinstance(probe, MetaPartition):
        result_dfs = [mp.data for mp in result]
    else:
        result_dfs = result
    df = pd.concat(result_dfs).reset_index(drop=True)

    # Assert column names are of type `str`, instead of `bytes` objects
    assert set(df.columns.map(type)) == {str}


def test_string_type_roundtrip(store_factory, bound_load_dataframes):
    # Note: we're not actually roundtripping the string type since the loading
    # type depends on the pandas version. Keeping the loading type aligned with
    # what is typically initialized by pandas by default is likely the best
    # option
    df = pd.DataFrame(
        {"str": pd.Series(["a", "b"], dtype="str" if PANDAS_3 else "string")}
    )

    store_dataframes_as_dataset(
        dfs=[df], store=store_factory, dataset_uuid="dataset_uuid"
    )

    result = bound_load_dataframes(dataset_uuid="dataset_uuid", store=store_factory)

    probe = result[0]
    if isinstance(probe, MetaPartition):
        result_dfs = [mp.data for mp in result]
    else:
        result_dfs = result
    result_df = pd.concat(result_dfs).reset_index(drop=True)
    pdt.assert_frame_equal(df, result_df)


def test_non_default_table_name_roundtrip(store_factory, bound_load_dataframes):
    df = pd.DataFrame({"A": [1]})
    store_dataframes_as_dataset(
        dfs=[df], store=store_factory, dataset_uuid="dataset_uuid", table_name="foo"
    )
    result = bound_load_dataframes(dataset_uuid="dataset_uuid", store=store_factory)

    probe = result[0]
    if isinstance(probe, MetaPartition):
        result_dfs = [mp.data for mp in result]
    else:
        result_dfs = result
    result_df = pd.concat(result_dfs).reset_index(drop=True)
    pdt.assert_frame_equal(df, result_df)



from collections.abc import Iterable
from functools import partial
from typing import cast

import pandas as pd
import pyarrow as pa
from minimalkv import KeyValueStore

from plateau.core import naming
from plateau.core.common_metadata import (
    SchemaWrapper,
    read_schema_metadata,
    store_schema_metadata,
    validate_compatible,
)
from plateau.core.dataset import DatasetMetadataBuilder
from plateau.core.factory import DatasetFactory
from plateau.core.index import ExplicitSecondaryIndex, IndexBase, PartitionIndex
from plateau.core.typing import StoreFactory, StoreInput
from plateau.core.utils import ensure_store
from plateau.io_components.metapartition import (
    SINGLE_TABLE,
    MetaPartition,
    MetaPartitionInput,
    parse_input_to_metapartition,
    partition_labels_from_mps,
)
from plateau.io_components.utils import (
    combine_metadata,
    extract_duplicates,
    sort_values_categorical,
)
from plateau.serialization import DataFrameSerializer

SINGLE_CATEGORY = SINGLE_TABLE


def write_partition(
    partition_df: MetaPartitionInput,
    secondary_indices: list[str],
    sort_partitions_by: list[str],
    dataset_uuid: str,
    partition_on: list[str],
    store_factory: StoreFactory,
    df_serializer: DataFrameSerializer | None,
    metadata_version: int,
    dataset_table_name: str = SINGLE_TABLE,
) -> MetaPartition:
    """Write a dataframe to store, performing all necessary preprocessing tasks
    like partitioning, bucketing (NotImplemented), indexing, etc.

    in the correct order.
    """
    store = ensure_store(store_factory)
    try:
        from plateau.io.dask._shuffle import _KTK_HASH_BUCKET

        if (
            isinstance(partition_df, pd.DataFrame)
            and _KTK_HASH_BUCKET in partition_df.columns
        ):
            partition_df = partition_df.drop(columns=_KTK_HASH_BUCKET)
    except ModuleNotFoundError:
        pass
    # I don't have access to the group values
    mps = parse_input_to_metapartition(
        partition_df,
        metadata_version=metadata_version,
        table_name=dataset_table_name,
    )
    if sort_partitions_by:
        mps = mps.apply(partial(sort_values_categorical, columns=sort_partitions_by))
    if partition_on:
        mps = mps.partition_on(partition_on)
    if secondary_indices:
        mps = mps.build_indices(secondary_indices)
    return mps.store_dataframes(
        store=store, dataset_uuid=dataset_uuid, df_serializer=df_serializer
    )


def persist_indices(
    store: StoreInput, dataset_uuid: str, indices: dict[str, IndexBase]
) -> dict[str, str]:
    store = ensure_store(store)
    output_filenames = {}
    for column, index in indices.items():
        # backwards compat
        if isinstance(index, dict):
            legacy_storage_key = (
                f"{dataset_uuid}.{column}{naming.EXTERNAL_INDEX_SUFFIX}"
            )
            index = ExplicitSecondaryIndex(
                column=column, index_dct=index, index_storage_key=legacy_storage_key
            )
        elif isinstance(index, PartitionIndex):
            continue
        index = cast(ExplicitSecondaryIndex, index)
        output_filenames[column] = index.store(store=store, dataset_uuid=dataset_uuid)
    return output_filenames


def persist_common_metadata(
    schemas: Iterable[SchemaWrapper],
    update_dataset: DatasetFactory | None,
    store: KeyValueStore,
    dataset_uuid: str,
    table_name: str,
):
    if not schemas:
        return None
    schemas_set = set(schemas)
    del schemas

    if update_dataset:
        schemas_set.add(
            read_schema_metadata(
                dataset_uuid=dataset_uuid, store=store, table=table_name
            )
        )

    schemas_sorted = sorted(schemas_set, key=lambda s: sorted(s.origin))

    try:
        result = validate_compatible(schemas_sorted)
    except ValueError as e:
        raise ValueError(
            f"Schemas for dataset '{dataset_uuid}' are not compatible!\n\n{e}"
        ) from e
    if result:
        store_schema_metadata(
            schema=result, dataset_uuid=dataset_uuid, store=store, table=table_name
        )
    return result


# Currently we only support nanosecond timestamps.
def coerce_schema_timestamps(wrapper: SchemaWrapper) -> SchemaWrapper:
    schema = wrapper.internal()
    fields = []
    for field in schema:
        if field.type in [pa.timestamp("s"), pa.timestamp("ms"), pa.timestamp("us")]:
            field = pa.field(field.name, pa.timestamp("ns"))
        fields.append(field)
    return SchemaWrapper(pa.schema(fields, schema.metadata), wrapper.origin)


def store_dataset_from_partitions(
    partition_list,
    store: StoreInput,
    dataset_uuid,
    dataset_metadata=None,
    metadata_merger=None,
    update_dataset=None,
    remove_partitions=None,
    metadata_storage_format=naming.DEFAULT_METADATA_STORAGE_FORMAT,
):
    store = ensure_store(store)

    schemas = set()
    if update_dataset:
        dataset_builder = DatasetMetadataBuilder.from_dataset(update_dataset)
        metadata_version = dataset_builder.metadata_version
        table_name = update_dataset.table_name
        schemas.add(update_dataset.schema)
    else:
        mp = next(iter(partition_list), None)

        if mp is None:
            raise ValueError(
                "Cannot store empty datasets, partition_list must not be empty if in store mode."
            )
        table_name = mp.table_name
        metadata_version = mp.metadata_version
        dataset_builder = DatasetMetadataBuilder(
            uuid=dataset_uuid,
            metadata_version=metadata_version,
            partition_keys=mp.partition_keys,
        )

    for mp in partition_list:
        if mp.schema:
            schemas.add(coerce_schema_timestamps(mp.schema))

    dataset_builder.schema = persist_common_metadata(
        schemas=schemas,
        update_dataset=update_dataset,
        store=store,
        dataset_uuid=dataset_uuid,
        table_name=table_name,
    )

    # We can only check for non unique partition labels here and if they occur we will
    # fail hard. The resulting dataset may be corrupted or file may be left in the store
    # without dataset metadata
    partition_labels = partition_labels_from_mps(partition_list)

    # This could be safely removed since we do not allow to set this by the user
    # anymore. It has implications on tests if mocks are used
    non_unique_labels = extract_duplicates(partition_labels)

    if non_unique_labels:
        raise ValueError(
            "The labels {} are duplicated. Dataset metadata was not written.".format(
                ", ".join(non_unique_labels)
            )
        )

    if remove_partitions is None:
        remove_partitions = []

    if metadata_merger is None:
        metadata_merger = combine_metadata

    dataset_builder = update_metadata(
        dataset_builder, metadata_merger, dataset_metadata
    )
    dataset_builder = update_partitions(
        dataset_builder, partition_list, remove_partitions
    )
    dataset_builder = update_indices(
        dataset_builder, store, partition_list, remove_partitions
    )
    if metadata_storage_format.lower() == "json":
        store.put(*dataset_builder.to_json())
    elif metadata_storage_format.lower() == "msgpack":
        store.put(*dataset_builder.to_msgpack())
    else:
        raise ValueError(
            f"Unknown metadata storage format encountered: {metadata_storage_format}"
        )
    dataset = dataset_builder.to_dataset()
    return dataset


def update_metadata(dataset_builder, metadata_merger, dataset_metadata):
    metadata_list = [dataset_builder.metadata]
    new_dataset_metadata = metadata_merger(metadata_list)

    dataset_metadata = dataset_metadata or {}

    if callable(dataset_metadata):
        dataset_metadata = dataset_metadata()

    new_dataset_metadata.update(dataset_metadata)
    for key, value in new_dataset_metadata.items():
        dataset_builder.add_metadata(key, value)
    return dataset_builder


def update_partitions(dataset_builder, add_partitions, remove_partitions):
    for mp in add_partitions:
        for mmp in mp:
            if mmp.label is not None:
                dataset_builder.explicit_partitions = True
                dataset_builder.add_partition(mmp.label, mmp.partition)

    for partition_name in remove_partitions:
        del dataset_builder.partitions[partition_name]

    return dataset_builder


def update_indices(dataset_builder, store, add_partitions, remove_partitions):
    dataset_indices = dataset_builder.indices
    partition_indices = MetaPartition.merge_indices(add_partitions)

    if dataset_indices:  # dataset already exists and will be updated
        if remove_partitions:
            for column, dataset_index in dataset_indices.items():
                dataset_indices[column] = dataset_index.remove_partitions(
                    remove_partitions, inplace=True
                )

        for column, index in partition_indices.items():
            dataset_indices[column] = dataset_indices[column].update(
                index, inplace=True
            )

    else:  # dataset index will be created first time from partitions
        dataset_indices = partition_indices

    # Store indices
    index_filenames = persist_indices(
        store=store, dataset_uuid=dataset_builder.uuid, indices=dataset_indices
    )
    for column, filename in index_filenames.items():
        dataset_builder.add_external_index(column, filename)

    return dataset_builder


def raise_if_dataset_exists(dataset_uuid, store):
    try:
        store_instance = ensure_store(store)
        for form in ["msgpack", "json"]:
            key = naming.metadata_key_from_uuid(uuid=dataset_uuid, format=form)
            if key in store_instance:
                raise RuntimeError(
                    "Dataset `%s` already exists and overwrite is not permitted!",
                    dataset_uuid,
                )
    except KeyError:
        pass



import pandas as pd
import pandas.testing as pdt

from plateau.io.dask.dataframe import hash_dataset


def test_hash_dataset(dataset_with_index_factory):
    hh = (
        hash_dataset(factory=dataset_with_index_factory)
        .compute()
        .reset_index(drop=True)
    )

    expected = pd.Series([11462879952839863487, 12568779102514529673], dtype="uint64")
    assert len(hh) == len(dataset_with_index_factory.partitions)
    pdt.assert_series_equal(hh, expected)


def test_hash_dataset_subset(dataset_with_index_factory):
    hh = (
        hash_dataset(factory=dataset_with_index_factory, subset=["TARGET"])
        .compute()
        .reset_index(drop=True)
    )

    expected = pd.Series([11358988112447789330, 826468140851422801], dtype="uint64")
    assert len(hh) == len(dataset_with_index_factory.partitions)
    pdt.assert_series_equal(hh, expected)


def test_hash_dataset_group_keys(dataset_with_index_factory):
    group_keys = ["L"]
    hh = (
        hash_dataset(factory=dataset_with_index_factory, group_key=group_keys)
        .compute()
        .sort_index()
    )

    expected = pd.Series(
        [11462879952839863487, 12568779102514529673],
        dtype="uint64",
        index=pd.Index([1, 2], name="L"),
    )
    pdt.assert_series_equal(hh, expected)


def test_hash_dataset_group_keys_subset(dataset_with_index_factory):
    group_keys = ["P"]
    hh = (
        hash_dataset(
            factory=dataset_with_index_factory, group_key=group_keys, subset=["TARGET"]
        )
        .compute()
        .sort_index()
    )

    expected = pd.Series(
        [11358988112447789330, 826468140851422801],
        index=pd.Index([1, 2], name="P"),
        dtype="uint64",
    )
    pdt.assert_series_equal(hh, expected)


def test_hash_dataset_group_keys_subset_subset_groupkey(dataset_with_index_factory):
    group_keys = ["P"]
    hh = (
        hash_dataset(
            factory=dataset_with_index_factory,
            group_key=group_keys,
            subset=["P", "TARGET"],
        )
        .compute()
        .sort_index()
    )

    expected = pd.Series(
        [7554402398462747209, 1687604933839263903],
        index=pd.Index([1, 2], name="P"),
        dtype="uint64",
    )
    pdt.assert_series_equal(hh, expected)



import random
from collections.abc import Callable, Iterable, Mapping, Sequence
from typing import (
    SupportsFloat,
    cast,
)

import dask
import dask.dataframe as dd
import numpy as np
import pandas as pd

from plateau.core.common_metadata import empty_dataframe_from_schema
from plateau.core.docs import default_docs
from plateau.core.factory import DatasetFactory, _ensure_factory
from plateau.core.naming import DEFAULT_METADATA_VERSION
from plateau.core.typing import StoreFactory, StoreInput
from plateau.io.dask.compression import pack_payload, unpack_payload_pandas
from plateau.io_components.metapartition import (
    _METADATA_SCHEMA,
    SINGLE_TABLE,
    MetaPartition,
    parse_input_to_metapartition,
)
from plateau.io_components.read import dispatch_metapartitions_from_factory
from plateau.io_components.update import update_dataset_from_partitions
from plateau.io_components.utils import (
    _ensure_compatible_indices,
    normalize_args,
    validate_partition_keys,
)
from plateau.io_components.write import (
    raise_if_dataset_exists,
    store_dataset_from_partitions,
    write_partition,
)
from plateau.serialization import DataFrameSerializer, PredicatesType

from ._shuffle import shuffle_store_dask_partitions
from ._utils import _maybe_get_categoricals_from_index

__all__ = (
    "read_dataset_as_ddf",
    "store_dataset_from_ddf",
    "update_dataset_from_ddf",
    "collect_dataset_metadata",
    "hash_dataset",
)
from dask.dataframe import from_map
from dask.dataframe.io.utils import DataFrameIOFunction


class ReadPlateauPartition(DataFrameIOFunction):
    def __init__(
        self,
        columns,
    ) -> None:
        self._columns = columns
        super().__init__()

    @property
    def columns(self):
        """Return the current column projection."""
        return self._columns

    def project_columns(self, columns):
        """Return a new DataFrameIOFunction object with a new column
        projection."""
        return type(self)(columns=columns)

    def __call__(self, mps, *args, columns: Sequence[str] | None = None, **kwargs):
        """Return a new DataFrame partition."""
        cols = (
            self.columns
            if columns is None
            else [c for c in columns if c in self.columns]
        )
        return MetaPartition.concat_metapartitions(
            [mp.load_dataframes(*args, columns=cols, **kwargs) for mp in mps]
        ).data


@default_docs
@normalize_args
def read_dataset_as_ddf(
    dataset_uuid=None,
    store=None,
    table=SINGLE_TABLE,
    columns=None,
    predicate_pushdown_to_io=True,
    categoricals: Sequence[str] | None = None,
    dates_as_object: bool = True,
    predicates=None,
    factory=None,
    dask_index_on=None,
    dispatch_by=None,
):
    """Retrieve a single table from a dataset as partition-individual
    :class:`~dask.dataframe.DataFrame` instance.

    Please take care when using categoricals with Dask. For index columns, this function will construct dataset
    wide categoricals. For all other columns, Dask will determine the categories on a partition level and will
    need to merge them when shuffling data.

    Parameters
    ----------
    dask_index_on: str
        Reconstruct (and set) a dask index on the provided index column. Cannot be used
        in conjunction with `dispatch_by`.

        For details on performance, see also `dispatch_by`
    """
    if dask_index_on is not None and not isinstance(dask_index_on, str):
        raise TypeError(
            f"The parameter `dask_index_on` must be a string but got {type(dask_index_on)}"
        )

    if dask_index_on is not None and dispatch_by is not None and len(dispatch_by) > 0:
        raise ValueError(
            "`read_dataset_as_ddf` got parameters `dask_index_on` and `dispatch_by`. "
            "Note that `dispatch_by` can only be used if `dask_index_on` is None."
        )

    ds_factory = _ensure_factory(
        dataset_uuid=dataset_uuid,
        store=store,
        factory=factory,
    )

    if isinstance(columns, dict):
        columns = columns[table]
    meta = _get_dask_meta_for_dataset(
        ds_factory, columns, categoricals, dates_as_object
    )

    if columns is None:
        columns = list(meta.columns)
    mps = list(
        dispatch_metapartitions_from_factory(
            dataset_factory=ds_factory,
            predicates=predicates,
            dispatch_by=dispatch_by or dask_index_on,  # type: ignore
        )
    )
    divisions_lst = None
    if dask_index_on:
        ds_factory.load_index(dask_index_on)
        divisions = ds_factory.indices[dask_index_on].observed_values()
        divisions_lst = list(divisions)
        divisions_lst = sorted(divisions_lst)
        divisions_lst.append(divisions_lst[-1])

    with dask.config.set({"dataframe.convert-string": False}):
        ddf = from_map(
            ReadPlateauPartition(columns=columns),
            mps,
            meta=meta,
            columns=columns,
            label="read-plateau",
            divisions=divisions_lst,
            store=ds_factory.store_factory,
            categoricals=categoricals,
            predicate_pushdown_to_io=predicate_pushdown_to_io,
            dates_as_object=dates_as_object,
            predicates=predicates,
        )
    if dask_index_on:
        return ddf.set_index(dask_index_on, divisions=divisions_lst, sorted=True)
    else:
        return ddf


def _get_dask_meta_for_dataset(ds_factory, columns, categoricals, dates_as_object):
    """Calculate a schema suitable for the dask dataframe meta from the
    dataset."""
    table_schema = ds_factory.schema
    meta = empty_dataframe_from_schema(
        table_schema, columns=columns, date_as_object=dates_as_object
    )

    if categoricals:
        meta = meta.astype(dict.fromkeys(categoricals, "category"))
        meta = dd.utils.clear_known_categories(meta, categoricals)

    categoricals_from_index = _maybe_get_categoricals_from_index(
        ds_factory, categoricals
    )
    if categoricals_from_index:
        meta = meta.astype(categoricals_from_index)
    return meta


def _shuffle_docs(func):
    func.__doc__ += """

    .. admonition:: Behavior without ``shuffle==False``

        In the case without ``partition_on`` every dask partition is mapped to a single plateau partition

        In the case with ``partition_on`` every dask partition is mapped to N plateau partitions, where N
        depends on the content of the respective partition, such that every resulting plateau partition has
        only a single value in the respective ``partition_on`` columns.

    .. admonition:: Behavior with ``shuffle==True``

        ``partition_on`` is mandatory

        Perform a data shuffle to ensure that every primary key will have at most ``num_bucket``.

        .. note::
            The number of allowed buckets will have an impact on the required resources and runtime.
            Using a larger number of allowed buckets will usually reduce resource consumption and in some
            cases also improves runtime performance.

        :Example:

            >>> partition_on="primary_key"
            >>> num_buckets=2  # doctest: +SKIP
            primary_key=1/bucket1.parquet
            primary_key=1/bucket2.parquet

    .. note:: This can only be used for datasets with a single table!

    See also, :ref:`partitioning_dask`.

    Parameters
    ----------
    ddf: Union[dask.dataframe.DataFrame, None]
        The dask.Dataframe to be used to calculate the new partitions from. If this parameter is `None`, the update pipeline
        will only delete partitions without creating new ones.
    shuffle: bool
        If `True` and `partition_on` is requested, shuffle the data to reduce number of output partitions.

        See also, :ref:`shuffling`.

        .. warning::

            Dask uses a heuristic to determine how data is shuffled and there are two options, `partd` for local disk shuffling and `tasks` for distributed shuffling using a task graph. If there is no :class:`distributed.Client` in the context and the option is not set explicitly, dask will choose `partd` which may cause data loss when the graph is executed on a distributed cluster.

            Therefore, we recommend to specify the dask shuffle method explicitly, e.g. by using a context manager.

            .. code::

                with dask.config.set(shuffle='tasks'):
                    graph = update_dataset_from_ddf(...)
                graph.compute()

    repartition_ratio: Optional[Union[int, float]]
        If provided, repartition the dataframe before calculation starts to ``ceil(ddf.npartitions / repartition_ratio)``
    num_buckets: int
        If provided, the output partitioning will have ``num_buckets`` files per primary key partitioning.
        This effectively splits up the execution ``num_buckets`` times. Setting this parameter may be helpful when
        scaling.
        This only has an effect if ``shuffle==True``
    bucket_by:
        The subset of columns which should be considered for bucketing.

        This parameter ensures that groups of the given subset are never split
        across buckets within a given partition.

        Without specifying this the buckets will be created randomly.

        This only has an effect if ``shuffle==True``

        .. admonition:: Secondary indices

            This parameter has a strong effect on the performance of secondary
            indices. Since it guarantees that a given tuple of the subset will
            be entirely put into the same file you can build efficient indices
            with this approach.

        .. note::

            Only columns with data types which can be hashed are allowed to be used in this.
"""
    return func


def _id(x):
    return x


def _commit_update_from_reduction(df_mps, **kwargs):
    partitions = pd.Series(
        filter(
            lambda mp: mp is not np.nan and not mp.is_sentinel, df_mps.values.flatten()
        )
    ).dropna()
    return update_dataset_from_partitions(
        partition_list=partitions,
        **kwargs,
    )


def _commit_store_from_reduction(df_mps, **kwargs):
    partitions = pd.Series(
        filter(
            lambda mp: mp is not np.nan and not mp.is_sentinel, df_mps.values.flatten()
        )
    ).dropna()
    return store_dataset_from_partitions(
        partition_list=partitions,
        **kwargs,
    )


@default_docs
@_shuffle_docs
@normalize_args
def store_dataset_from_ddf(
    ddf: dd.DataFrame,
    store: StoreInput,
    dataset_uuid: str,
    table: str = SINGLE_TABLE,
    secondary_indices: list[str] | None = None,
    shuffle: bool = False,
    repartition_ratio: SupportsFloat | None = None,
    num_buckets: int = 1,
    sort_partitions_by: list[str] | str | None = None,
    metadata: Mapping | None = None,
    df_serializer: DataFrameSerializer | None = None,
    metadata_merger: Callable | None = None,
    metadata_version: int = DEFAULT_METADATA_VERSION,
    partition_on: list[str] | None = None,
    bucket_by: list[str] | str | None = None,
    overwrite: bool = False,
):
    """Store a dataset from a dask.dataframe."""
    # normalization done by normalize_args but mypy doesn't recognize this
    sort_partitions_by = cast(list[str], sort_partitions_by)
    secondary_indices = cast(list[str], secondary_indices)
    bucket_by = cast(list[str], bucket_by)
    partition_on = cast(list[str], partition_on)

    if table is None:
        raise TypeError("The parameter `table` is not optional.")

    ds_factory = _ensure_factory(dataset_uuid=dataset_uuid, store=store, factory=None)

    if not overwrite:
        raise_if_dataset_exists(dataset_uuid=dataset_uuid, store=store)

    with dask.config.set(
        {"dataframe.convert-string": False, "dataframe.shuffle.method": "tasks"}
    ):
        mp_ser = _write_dataframe_partitions(
            ddf=ddf,
            store=ds_factory.store_factory,
            dataset_uuid=dataset_uuid,
            table=table,
            secondary_indices=secondary_indices,
            shuffle=shuffle,
            repartition_ratio=repartition_ratio,
            num_buckets=num_buckets,
            sort_partitions_by=sort_partitions_by,
            df_serializer=df_serializer,
            metadata_version=metadata_version,
            partition_on=partition_on,
            bucket_by=bucket_by,
        )

    return mp_ser.reduction(
        chunk=_id,
        aggregate=_commit_store_from_reduction,
        split_every=False,
        token="commit-dataset",
        meta=object,
        aggregate_kwargs={
            "store": ds_factory.store_factory,
            "dataset_uuid": ds_factory.dataset_uuid,
            "dataset_metadata": metadata,
            "metadata_merger": metadata_merger,
        },
    )


def _write_dataframe_partitions(
    ddf: dd.DataFrame,
    store: StoreFactory,
    dataset_uuid: str,
    table: str,
    secondary_indices: list[str],
    shuffle: bool,
    repartition_ratio: SupportsFloat | None,
    num_buckets: int,
    sort_partitions_by: list[str],
    df_serializer: DataFrameSerializer | None,
    metadata_version: int,
    partition_on: list[str],
    bucket_by: list[str],
) -> dd.Series:
    if repartition_ratio and ddf is not None:
        ddf = ddf.repartition(
            npartitions=int(np.ceil(ddf.npartitions / repartition_ratio))
        )

    if ddf is None:
        mps = dd.from_pandas(
            pd.Series(
                [
                    parse_input_to_metapartition(
                        None,
                        metadata_version=metadata_version,
                        table_name=table,
                    )
                ]
            ),
            npartitions=1,
        )
    else:
        if shuffle:
            mps = shuffle_store_dask_partitions(
                ddf=ddf,
                table=table,
                secondary_indices=secondary_indices,
                metadata_version=metadata_version,
                partition_on=partition_on,
                store_factory=store,
                df_serializer=df_serializer,
                dataset_uuid=dataset_uuid,
                num_buckets=num_buckets,
                sort_partitions_by=sort_partitions_by,
                bucket_by=bucket_by,
            )
        else:
            mps = ddf.map_partitions(
                write_partition,
                secondary_indices=secondary_indices,
                metadata_version=metadata_version,
                partition_on=partition_on,
                store_factory=store,
                df_serializer=df_serializer,
                dataset_uuid=dataset_uuid,
                sort_partitions_by=sort_partitions_by,
                dataset_table_name=table,
                meta=(MetaPartition),
            )
    return mps


@default_docs
@_shuffle_docs
@normalize_args
def update_dataset_from_ddf(
    ddf: dd.DataFrame,
    store: StoreInput | None = None,
    dataset_uuid: str | None = None,
    table: str = SINGLE_TABLE,
    secondary_indices: list[str] | None = None,
    shuffle: bool = False,
    repartition_ratio: SupportsFloat | None = None,
    num_buckets: int = 1,
    sort_partitions_by: list[str] | str | None = None,
    delete_scope: Iterable[Mapping[str, str]] | None = None,
    metadata: Mapping | None = None,
    df_serializer: DataFrameSerializer | None = None,
    metadata_merger: Callable | None = None,
    default_metadata_version: int = DEFAULT_METADATA_VERSION,
    partition_on: list[str] | None = None,
    factory: DatasetFactory | None = None,
    bucket_by: list[str] | str | None = None,
):
    """Update a dataset from a dask.dataframe.

    See Also
    --------
    :ref:`mutating_datasets`
    """
    if table is None:
        raise TypeError("The parameter `table` is not optional.")

    # normalization done by normalize_args but mypy doesn't recognize this
    sort_partitions_by = cast(list[str], sort_partitions_by)
    secondary_indices = cast(list[str], secondary_indices)
    bucket_by = cast(list[str], bucket_by)
    partition_on = cast(list[str], partition_on)

    ds_factory, metadata_version, partition_on = validate_partition_keys(
        dataset_uuid=dataset_uuid,
        store=store,
        default_metadata_version=default_metadata_version,
        partition_on=partition_on,
        ds_factory=factory,
    )

    inferred_indices = _ensure_compatible_indices(ds_factory, secondary_indices)
    del secondary_indices

    if ds_factory:
        store_factory: StoreFactory = ds_factory.store_factory
    else:
        if not callable(store):
            raise TypeError(
                "You must either pass in a DatasetFactory or a StoreFactor via store."
            )
        store_factory = store

    with dask.config.set(
        {"dataframe.convert-string": False, "dataframe.shuffle.method": "tasks"}
    ):
        mp_ser = _write_dataframe_partitions(
            ddf=ddf,
            store=store_factory,
            dataset_uuid=dataset_uuid or ds_factory.dataset_uuid,
            table=table,
            secondary_indices=inferred_indices,
            shuffle=shuffle,
            repartition_ratio=repartition_ratio,
            num_buckets=num_buckets,
            sort_partitions_by=sort_partitions_by,
            df_serializer=df_serializer,
            metadata_version=metadata_version,
            partition_on=cast(list[str], partition_on),
            bucket_by=bucket_by,
        )

    final = mp_ser.reduction(
        chunk=_id,
        aggregate=_commit_update_from_reduction,
        split_every=False,
        token="commit-dataset",
        meta=object,
        aggregate_kwargs={
            "store_factory": store,
            "dataset_uuid": dataset_uuid,
            "ds_factory": ds_factory,
            "delete_scope": delete_scope,
            "metadata": metadata,
            "metadata_merger": metadata_merger,
        },
    )
    return final


@default_docs
@normalize_args
def collect_dataset_metadata(
    store: StoreInput | None = None,
    dataset_uuid: str | None = None,
    predicates: PredicatesType | None = None,
    frac: float = 1.0,
    factory: DatasetFactory | None = None,
) -> dd.DataFrame:
    """Collect parquet metadata of the dataset. The `frac` parameter can be
    used to select a subset of the data.

    .. warning::
      If the size of the partitions is not evenly distributed, e.g. some partitions might be larger than others,
      the metadata returned is not a good approximation for the whole dataset metadata.
    .. warning::
      Using the `frac` parameter is not encouraged for a small number of total partitions.


    Parameters
    ----------
    predicates
      plateau predicates to apply filters on the data for which to gather statistics

      .. warning::
          Filtering will only be applied for predicates on indices.
          The evaluation of the predicates therefore will therefore only return an approximate result.

    frac
      Fraction of the total number of partitions to use for gathering statistics. `frac == 1.0` will use all partitions.

    Returns
    -------
    dask.dataframe.DataFrame:
        A dask.DataFrame containing the following information about dataset statistics:
        * `partition_label`: File name of the parquet file, unique to each physical partition.
        * `row_group_id`: Index of the row groups within one parquet file.
        * `row_group_compressed_size`: Byte size of the data within one row group.
        * `row_group_uncompressed_size`: Byte size (uncompressed) of the data within one row group.
        * `number_rows_total`: Total number of rows in one parquet file.
        * `number_row_groups`: Number of row groups in one parquet file.
        * `serialized_size`: Serialized size of the parquet file.
        * `number_rows_per_row_group`: Number of rows per row group.

    Raises
    ------
    ValueError
      If no metadata could be retrieved, raise an error.
    """
    if not 0.0 < frac <= 1.0:
        raise ValueError(
            f"Invalid value for parameter `frac`: {frac}."
            "Please make sure to provide a value larger than 0.0 and smaller than or equal to 1.0 ."
        )
    dataset_factory = _ensure_factory(
        dataset_uuid=dataset_uuid,
        store=store,
        factory=factory,
    )

    mps = list(
        dispatch_metapartitions_from_factory(dataset_factory, predicates=predicates)
    )

    with dask.config.set({"dataframe.convert-string": False}):
        if mps:
            random.shuffle(mps)
            # ensure that even with sampling at least one metapartition is returned
            cutoff_index = max(1, int(len(mps) * frac))
            mps = mps[:cutoff_index]
            ddf = dd.from_map(
                MetaPartition.get_parquet_metadata,
                mps,
                meta=_METADATA_SCHEMA,
                store=dataset_factory.store_factory,
            )
        else:
            df = pd.DataFrame(columns=_METADATA_SCHEMA.keys())
            df = df.astype(_METADATA_SCHEMA)
            ddf = dd.from_pandas(df, npartitions=1)

    return ddf


def _unpack_hash(df, group_key, unpack_meta, subset):
    df = unpack_payload_pandas(df, unpack_meta).set_index(group_key, drop=True)
    return df.groupby(df.index).apply(_hash_partition, subset=subset)


def _hash_partition(part, subset):
    return pd.util.hash_pandas_object(part.reset_index()[subset], index=False).sum()


@default_docs
@normalize_args
def hash_dataset(
    store: StoreInput | None = None,
    dataset_uuid: str | None = None,
    subset=None,
    group_key=None,
    table: str = SINGLE_TABLE,
    predicates: PredicatesType | None = None,
    factory: DatasetFactory | None = None,
) -> dd.Series:
    """Calculate a partition wise, or group wise, hash of the dataset.

    .. note::

        We do not guarantee the hash values to remain constant across versions.


    Example output::

        Assuming a dataset with two unique values in column `P` this gives

        >>> hash_dataset(factory=dataset_with_index_factory,group_key=["P"]).compute()
        ... P
        ... 1    11462879952839863487
        ... 2    12568779102514529673
        ... dtype: uint64

    Parameters
    ----------
    subset
        If provided, only take these columns into account when hashing the dataset
    group_key
        If provided, calculate hash per group instead of per partition
    """
    dataset_factory = _ensure_factory(
        dataset_uuid=dataset_uuid,
        store=store,
        factory=factory,
    )

    columns = subset
    if subset and group_key:
        columns = sorted(set(subset) | set(group_key))
    ddf = read_dataset_as_ddf(
        table=table,
        predicates=predicates,
        factory=dataset_factory,
        columns=columns,
        dates_as_object=True,
    )
    with dask.config.set(
        {"dataframe.convert-string": False, "dataframe.shuffle.method": "tasks"}
    ):
        subset = subset or ddf.columns.to_list()
        if not group_key:
            return ddf.map_partitions(
                _hash_partition, subset=subset, meta=(None, "uint64")
            )
        else:
            ddf2 = pack_payload(ddf, group_key=group_key)
            final = ddf2.shuffle(on=group_key).map_partitions(
                _unpack_hash,
                group_key=group_key,
                unpack_meta=ddf._meta,
                subset=subset,
                meta=(None, "uint64"),
            )

            return final



from datetime import date, datetime

import numpy as np
import pandas as pd


def _to_binary(x):
    return str(x).encode("utf-8")


BINARY_COLUMNS = [
    _to_binary(chr(1)),
    b"1",
    b"2",
    b"3",
    "4".encode("utf-16"),
    "4".encode("utf-32"),
    # this is a type1 UUID
    b"\x8f\xb6\xe5@\x90\xdc\x11\xe8\xa0\xae\x02B\xac\x12\x01\x06",
    "🙈".encode(),
    _to_binary(chr(128)),
]


def get_dataframe_not_nested(n):
    if n > len(BINARY_COLUMNS):
        n_gen = n - len(BINARY_COLUMNS)
        binaries = BINARY_COLUMNS + [
            _to_binary(x)
            for x in range(len(BINARY_COLUMNS), n_gen + len(BINARY_COLUMNS))
        ]
    else:
        binaries = BINARY_COLUMNS[:n]
    return pd.DataFrame(
        {
            "bool": pd.Series(
                [1] * int(np.floor(n / 2)) + [0] * int(np.ceil(n / 2)), dtype=np.bool_
            ),
            "int8": pd.Series(range(n)).astype(np.int8),
            "int16": pd.Series(range(n)).astype(np.int16),
            "int32": pd.Series(range(n)).astype(np.int32),
            "int64": pd.Series(range(n)).astype(np.int64),
            "uint8": pd.Series(range(n)).astype(np.uint8),
            "uint16": pd.Series(range(n)).astype(np.uint16),
            "uint32": pd.Series(range(n)).astype(np.uint32),
            "uint64": pd.Series(range(n)).astype(np.uint64),
            "float32": pd.Series([float(x) for x in range(n)]).astype(np.float32),
            "float64": pd.Series([float(x) for x in range(n)]).astype(np.float64),
            "date": pd.Series(
                [date(2018, 1, x % 31 + 1) for x in range(1, n + 1)], dtype=object
            ),
            "datetime64": pd.Series(
                [datetime(2018, 1, x % 31 + 1) for x in range(1, n + 1)],
                dtype="datetime64[ns]",
            ),
            "unicode": pd.Series([str(x) for x in range(n)], dtype=str),
            "null": pd.Series([None] * n, dtype=object),
            "bytes": pd.Series(binaries, dtype=object),
        }
    ).sort_index(axis=1)



import numpy as np
import pandas as pd
import pytest
from toolz.dicttoolz import valmap

from plateau.core.factory import DatasetFactory
from plateau.io.eager import store_dataframes_as_dataset


def assert_index_dct_equal(dict1, dict2):
    dict1 = valmap(sorted, dict1)
    dict2 = valmap(sorted, dict2)
    assert dict1 == dict2


def test_build_indices(store_factory, metadata_version, bound_build_dataset_indices):
    dataset_uuid = "dataset_uuid"
    partitions = [
        pd.DataFrame({"p": [1, 2]}),
        pd.DataFrame({"p": [2, 3]}),
    ]

    dataset = store_dataframes_as_dataset(
        dfs=partitions,
        store=store_factory,
        dataset_uuid=dataset_uuid,
        metadata_version=metadata_version,
    )
    dataset = dataset.load_all_indices(store=store_factory)
    assert not dataset.indices

    # Create indices
    bound_build_dataset_indices(store_factory, dataset_uuid, columns=["p"])

    # Assert indices are properly created
    dataset_factory = DatasetFactory(dataset_uuid, store_factory, load_all_indices=True)
    index_dct = dataset_factory.indices["p"].index_dct

    assert len(index_dct[1]) == 1
    assert len(index_dct[2]) == 2
    assert len(index_dct[3]) == 1

    assert len(set(index_dct[3]) & set(index_dct[2])) == 1
    assert len(set(index_dct[1]) & set(index_dct[2])) == 1
    assert len(set(index_dct[1]) & set(index_dct[3])) == 0


def test_create_index_from_inexistent_column_fails(
    store_factory, metadata_version, bound_build_dataset_indices
):
    dataset_uuid = "dataset_uuid"
    partitions = [
        pd.DataFrame({"p": [1, 2]}),
        pd.DataFrame({"p": [2, 3]}),
    ]

    store_dataframes_as_dataset(
        dfs=partitions,
        store=store_factory,
        dataset_uuid=dataset_uuid,
        metadata_version=metadata_version,
    )

    with pytest.raises(RuntimeError, match="Column `.*` could not be found"):
        bound_build_dataset_indices(store_factory, dataset_uuid, columns=["abc"])


def test_add_column_to_existing_index(
    store_factory, metadata_version, bound_build_dataset_indices
):
    dataset_uuid = "dataset_uuid"
    partitions = [
        pd.DataFrame({"p": [1, 2], "x": [100, 4500]}),
        pd.DataFrame({"p": [4, 3], "x": [500, 10]}),
    ]

    dataset = store_dataframes_as_dataset(
        dfs=partitions,
        store=store_factory,
        dataset_uuid=dataset_uuid,
        metadata_version=metadata_version,
        secondary_indices="p",
    )
    assert dataset.load_all_indices(store=store_factory()).indices.keys() == {"p"}

    # Create indices
    bound_build_dataset_indices(store_factory, dataset_uuid, columns=["x"])

    # Assert indices are properly created
    dataset_factory = DatasetFactory(dataset_uuid, store_factory, load_all_indices=True)
    assert dataset_factory.indices.keys() == {"p", "x"}


def test_indices_uints(store_factory, metadata_version, bound_build_dataset_indices):
    dataset_uuid = "dataset_uuid"

    # min uint64
    p1 = 0

    # max uint64 => cannot even be cast to float32
    p2 = int(~np.uint64(0))

    # number that would be cut if converted to float64 and back
    p3 = 17128351978467489013

    partitions = [
        pd.DataFrame({"p": pd.Series([p1], dtype=np.uint64)}),
        pd.DataFrame({"p": pd.Series([p2], dtype=np.uint64)}),
        pd.DataFrame({"p": pd.Series([p3], dtype=np.uint64)}),
    ]

    def assert_expected(index_dct):
        assert len(index_dct) == 3
        referenced_partitions = []
        for val in index_dct.values():
            referenced_partitions.extend(val)
        assert len(referenced_partitions) == 3

    dataset = store_dataframes_as_dataset(
        dfs=partitions,
        store=store_factory,
        dataset_uuid=dataset_uuid,
        metadata_version=metadata_version,
    )
    dataset = dataset.load_all_indices(store=store_factory)
    assert not dataset.indices

    # Create indices
    bound_build_dataset_indices(store_factory, dataset_uuid, columns=["p"])

    # Assert indices are properly created
    dataset_factory = DatasetFactory(dataset_uuid, store_factory, load_all_indices=True)
    assert_expected(dataset_factory.indices["p"].index_dct)
    first_run = dataset_factory.indices["p"].index_dct.copy()

    # Re-create indices
    bound_build_dataset_indices(store_factory, dataset_uuid, columns=["p"])

    # Assert indices are properly created
    dataset_factory = DatasetFactory(dataset_uuid, store_factory, load_all_indices=True)
    assert_index_dct_equal(first_run, dataset_factory.indices["p"].index_dct)


def test_empty_partitions(store_factory, metadata_version, bound_build_dataset_indices):
    dataset_uuid = "dataset_uuid"

    partitions = [
        pd.DataFrame({"p": pd.Series([], dtype=np.int8)}),
        pd.DataFrame({"p": pd.Series([1], dtype=np.int8)}),
    ]

    dataset = store_dataframes_as_dataset(
        dfs=partitions,
        store=store_factory,
        dataset_uuid=dataset_uuid,
        metadata_version=metadata_version,
    )
    dataset = dataset.load_all_indices(store=store_factory)
    assert not dataset.indices

    # Create indices
    bound_build_dataset_indices(store_factory, dataset_uuid, columns=["p"])

    # Assert indices are properly created
    dataset_factory = DatasetFactory(dataset_uuid, store_factory, load_all_indices=True)
    assert len(dataset_factory.indices["p"].index_dct) == 1



import msgpack

try:
    import zstandard as zstd
except ImportError:
    # zstandard < 0.15.0
    import zstd  # type: ignore


def packb(obj):
    cctx = zstd.ZstdCompressor(write_content_size=True)
    compressed = cctx.compress(msgpack.packb(obj))
    return compressed


def unpackb(bts):
    dctx = zstd.ZstdDecompressor()
    decompressed = dctx.decompress(bts)
    return msgpack.unpackb(decompressed, raw=False)



"""This module tests high level dataset API functions which require entire
datasets, indices, etc."""

from collections import OrderedDict

import pandas as pd
import pandas.testing as pdt

from plateau.core._compat import pandas_infer_string
from plateau.core.dataset import DatasetMetadata
from plateau.core.index import ExplicitSecondaryIndex


def test_dataset_get_indices_as_dataframe_partition_keys_only(
    dataset_with_index, store_session
):
    expected = pd.DataFrame(
        {"P": [1, 2]},
        index=pd.Index(["P=1/cluster_1", "P=2/cluster_2"], name="partition"),
    )
    ds = dataset_with_index.load_partition_indices()
    result = ds.get_indices_as_dataframe(columns=dataset_with_index.partition_keys)
    pdt.assert_frame_equal(result, expected)


def test_dataset_get_indices_as_dataframe(dataset_with_index, store_session):
    expected = pd.DataFrame(
        OrderedDict([("L", [1, 2]), ("P", [1, 2])]),
        index=pd.Index(["P=1/cluster_1", "P=2/cluster_2"], name="partition"),
    )
    ds = dataset_with_index.load_partition_indices()
    ds = ds.load_index("L", store_session)

    result = ds.get_indices_as_dataframe()
    pdt.assert_frame_equal(result, expected)


def test_dataset_get_indices_as_dataframe_duplicates():
    ds = DatasetMetadata(
        "some_uuid",
        indices={
            "l_external_code": ExplicitSecondaryIndex(
                "l_external_code", {"1": ["part1", "part2"], "2": ["part1", "part2"]}
            ),
            "p_external_code": ExplicitSecondaryIndex(
                "p_external_code", {"1": ["part1"], "2": ["part2"]}
            ),
        },
    )
    expected = pd.DataFrame(
        OrderedDict(
            [
                ("p_external_code", ["1", "1", "2", "2"]),
                ("l_external_code", ["1", "2", "1", "2"]),
            ]
        ),
        index=pd.Index(["part1", "part1", "part2", "part2"], name="partition"),
    )
    result = ds.get_indices_as_dataframe()
    pdt.assert_frame_equal(result, expected)


def test_dataset_get_indices_as_dataframe_predicates():
    ds = DatasetMetadata(
        "some_uuid",
        indices={
            "l_external_code": ExplicitSecondaryIndex(
                "l_external_code", {"1": ["part1", "part2"], "2": ["part1", "part2"]}
            ),
            "p_external_code": ExplicitSecondaryIndex(
                "p_external_code", {"1": ["part1"], "2": ["part2"]}
            ),
        },
    )
    expected = pd.DataFrame(
        OrderedDict([("p_external_code", ["1"])]),
        index=pd.Index(["part1"], name="partition"),
    )
    result = ds.get_indices_as_dataframe(
        columns=["p_external_code"], predicates=[[("p_external_code", "==", "1")]]
    )
    pdt.assert_frame_equal(result, expected)

    result = ds.get_indices_as_dataframe(
        columns=["l_external_code"], predicates=[[("l_external_code", "==", "1")]]
    )
    expected = pd.DataFrame(
        OrderedDict([("l_external_code", "1")]),
        index=pd.Index(["part1", "part2"], name="partition"),
    )
    pdt.assert_frame_equal(result, expected)

    result = ds.get_indices_as_dataframe(
        columns=["l_external_code"],
        predicates=[[("l_external_code", "==", "1"), ("p_external_code", "==", "1")]],
    )
    expected = pd.DataFrame(
        OrderedDict([("l_external_code", "1")]),
        index=pd.Index(["part1"], name="partition"),
    )
    pdt.assert_frame_equal(result, expected)

    result = ds.get_indices_as_dataframe(
        columns=["l_external_code"],
        predicates=[[("l_external_code", "==", "1"), ("p_external_code", "==", "3")]],
    )
    empty_index = pd.Index([], name="partition")
    expected = pd.DataFrame(columns=["l_external_code"], index=empty_index)
    if pandas_infer_string():
        expected["l_external_code"] = expected["l_external_code"].astype(str)
    pdt.assert_frame_equal(result, expected)


def test_dataset_get_indices_as_dataframe_no_index(dataset):
    assert not dataset.primary_indices_loaded
    df = dataset.get_indices_as_dataframe()
    pdt.assert_frame_equal(df, pd.DataFrame(index=["cluster_1", "cluster_2"]))


def test_dataset_get_indices_as_dataframe_with_index(dataset_with_index, store_session):
    assert not dataset_with_index.primary_indices_loaded
    df = dataset_with_index.get_indices_as_dataframe()
    empty_index = pd.Index([], name="partition")
    pdt.assert_frame_equal(
        df,
        pd.DataFrame(columns=["L", "P"], index=empty_index).astype(
            {"L": "object", "P": "int64"}
        ),
    )



import subprocess

import pytest


@pytest.mark.parametrize(
    "module",
    [
        "plateau",
        "plateau.serialization",
        "plateau.io",
        "plateau.io_components",
        "plateau.utils",
        "plateau.api",
        "plateau.io.dask",
        "plateau.io.testing",
        "plateau.io.eager",
        "plateau.io.iter",
    ],
)
def test_imports(module):
    subprocess.run(["python", "-c", f"import {module}"], check=True)



from collections.abc import Sequence
from functools import partial

import dask
from dask import delayed
from dask.delayed import Delayed

from plateau.core import naming
from plateau.core.docs import default_docs
from plateau.core.factory import _ensure_factory
from plateau.core.naming import DEFAULT_METADATA_VERSION
from plateau.core.typing import StoreInput
from plateau.core.utils import lazy_store
from plateau.core.uuid import gen_uuid
from plateau.io_components.delete import (
    delete_common_metadata,
    delete_indices,
    delete_top_level_metadata,
)
from plateau.io_components.gc import delete_files, dispatch_files_to_gc
from plateau.io_components.metapartition import (
    SINGLE_TABLE,
    MetaPartition,
    parse_input_to_metapartition,
)
from plateau.io_components.read import dispatch_metapartitions_from_factory
from plateau.io_components.update import update_dataset_from_partitions
from plateau.io_components.utils import (
    _ensure_compatible_indices,
    normalize_arg,
    normalize_args,
    raise_if_indices_overlap,
    validate_partition_keys,
)
from plateau.io_components.write import (
    raise_if_dataset_exists,
    store_dataset_from_partitions,
    write_partition,
)

from ._utils import (
    _cast_categorical_to_index_cat,
    _get_data,
    _maybe_get_categoricals_from_index,
    map_delayed,
)

__all__ = (
    "delete_dataset__delayed",
    "garbage_collect_dataset__delayed",
    "read_dataset_as_delayed",
    "update_dataset_from_delayed",
    "store_delayed_as_dataset",
)


def _delete_all_additional_metadata(dataset_factory):
    delete_indices(dataset_factory=dataset_factory)
    delete_common_metadata(dataset_factory=dataset_factory)


def _delete_tl_metadata(dataset_factory, *args):
    """This function serves as a collector function for delayed objects.

    Therefore allowing additional arguments which are not used.
    """
    delete_top_level_metadata(dataset_factory=dataset_factory)


@default_docs
@normalize_args
def delete_dataset__delayed(dataset_uuid=None, store=None, factory=None):
    """
    Parameters
    ----------
    """
    dataset_factory = _ensure_factory(
        dataset_uuid=dataset_uuid,
        store=store,
        factory=factory,
        load_schema=False,
    )

    gc = garbage_collect_dataset__delayed(factory=dataset_factory)

    mps = dispatch_metapartitions_from_factory(dataset_factory)

    delayed_dataset_uuid = delayed(_delete_all_additional_metadata)(
        dataset_factory=dataset_factory
    )
    mps = map_delayed(
        MetaPartition.delete_from_store,
        mps,
        store=store,
        dataset_uuid=dataset_factory.dataset_uuid,
    )

    return delayed(_delete_tl_metadata)(
        dataset_factory, list(mps), gc, delayed_dataset_uuid
    )


@default_docs
@normalize_args
def garbage_collect_dataset__delayed(
    dataset_uuid: str | None = None,
    store: StoreInput | None = None,
    chunk_size: int = 100,
    factory=None,
) -> list[Delayed]:
    """Remove auxiliary files that are no longer tracked by the dataset.

    These files include indices that are no longer referenced by the metadata
    as well as files in the directories of the tables that are no longer
    referenced. The latter is only applied to static datasets.

    Parameters
    ----------
    chunk_size
        Number of files that should be deleted in a single job.
    """

    ds_factory = _ensure_factory(
        dataset_uuid=dataset_uuid,
        store=store,
        factory=factory,
    )

    nested_files = dispatch_files_to_gc(
        dataset_uuid=None, store_factory=None, chunk_size=chunk_size, factory=ds_factory
    )
    return list(
        map_delayed(delete_files, nested_files, store_factory=ds_factory.store_factory)
    )


def _load_and_concat_metapartitions_inner(mps, args, kwargs):
    return MetaPartition.concat_metapartitions(
        [mp.load_dataframes(*args, **kwargs) for mp in mps]
    )


def _load_and_concat_metapartitions(list_of_mps, *args, **kwargs):
    return map_delayed(
        _load_and_concat_metapartitions_inner, list_of_mps, args=args, kwargs=kwargs
    )


@default_docs
@normalize_args
def read_dataset_as_delayed_metapartitions(
    dataset_uuid=None,
    store=None,
    columns=None,
    predicate_pushdown_to_io=True,
    categoricals: Sequence[str] | None = None,
    dates_as_object: bool = True,
    predicates=None,
    factory=None,
    dispatch_by=None,
):
    """A collection of dask.delayed objects to retrieve a dataset from store
    where each partition is loaded as a
    :class:`~plateau.io_components.metapartition.MetaPartition`.

    .. seealso:

        :func:`~plateau.io.dask.read_dataset_as_delayed`

    Parameters
    ----------
    """
    ds_factory = _ensure_factory(
        dataset_uuid=dataset_uuid,
        store=store,
        factory=factory,
    )

    store = ds_factory.store_factory
    mps = dispatch_metapartitions_from_factory(
        dataset_factory=ds_factory,
        predicates=predicates,
        dispatch_by=dispatch_by,
    )

    if dispatch_by is not None:
        mps = _load_and_concat_metapartitions(
            mps,
            store=store,
            columns=columns,
            categoricals=categoricals,
            predicate_pushdown_to_io=predicate_pushdown_to_io,
            dates_as_object=dates_as_object,
            predicates=predicates,
        )
    else:
        mps = map_delayed(
            MetaPartition.load_dataframes,
            mps,
            store=store,
            columns=columns,
            categoricals=categoricals,
            predicate_pushdown_to_io=predicate_pushdown_to_io,
            dates_as_object=dates_as_object,
            predicates=predicates,
        )

    categoricals_from_index = _maybe_get_categoricals_from_index(
        ds_factory, categoricals
    )

    if categoricals_from_index:
        mps = map_delayed(
            partial(  # type: ignore
                MetaPartition.apply,
                func=partial(  # type: ignore
                    _cast_categorical_to_index_cat, categories=categoricals_from_index
                ),
                type_safe=True,
            ),
            mps,
        )

    return list(mps)


@default_docs
def read_dataset_as_delayed(
    dataset_uuid=None,
    store=None,
    columns=None,
    predicate_pushdown_to_io=True,
    categoricals=None,
    dates_as_object: bool = True,
    predicates=None,
    factory=None,
    dispatch_by=None,
):
    """A collection of dask.delayed objects to retrieve a dataset from store
    where each partition is loaded as a :class:`~pandas.DataFrame`.

    Parameters
    ----------
    """
    mps = read_dataset_as_delayed_metapartitions(
        dataset_uuid=dataset_uuid,
        store=store,
        factory=factory,
        columns=columns,
        predicate_pushdown_to_io=predicate_pushdown_to_io,
        categoricals=categoricals,
        dates_as_object=dates_as_object,
        predicates=predicates,
        dispatch_by=dispatch_by,
    )
    return list(map_delayed(_get_data, mps))


@default_docs
def update_dataset_from_delayed(
    delayed_tasks: list[Delayed],
    store=None,
    dataset_uuid=None,
    delete_scope=None,
    metadata=None,
    df_serializer=None,
    metadata_merger=None,
    default_metadata_version=DEFAULT_METADATA_VERSION,
    partition_on=None,
    sort_partitions_by=None,
    secondary_indices=None,
    factory=None,
    table_name=SINGLE_TABLE,
):
    """A dask.delayed graph to add and store a list of dictionaries containing
    dataframes to a plateau dataset in store. The input should be a list (or
    splitter pipeline) containing
    :class:`~plateau.io_components.metapartition.MetaPartition`. If you want to
    use this pipeline step for just deleting partitions without adding new ones
    you have to give an empty meta partition as input
    (``[Metapartition(None)]``).

    Parameters
    ----------

    See Also
    --------
    :ref:`mutating_datasets`
    """
    partition_on = normalize_arg("partition_on", partition_on)
    store = normalize_arg("store", store)
    secondary_indices = normalize_arg("secondary_indices", secondary_indices)
    delete_scope = dask.delayed(normalize_arg)("delete_scope", delete_scope)

    ds_factory, metadata_version, partition_on = validate_partition_keys(
        dataset_uuid=dataset_uuid,
        store=store,
        default_metadata_version=default_metadata_version,
        partition_on=partition_on,
        ds_factory=factory,
    )

    secondary_indices = _ensure_compatible_indices(ds_factory, secondary_indices)
    mps = map_delayed(
        write_partition,
        delayed_tasks,
        secondary_indices=secondary_indices,
        metadata_version=metadata_version,
        partition_on=partition_on,
        store_factory=store,
        df_serializer=df_serializer,
        dataset_uuid=dataset_uuid,
        sort_partitions_by=sort_partitions_by,
        dataset_table_name=table_name,
    )

    return dask.delayed(update_dataset_from_partitions)(
        list(mps),
        store_factory=store,
        dataset_uuid=dataset_uuid,
        ds_factory=ds_factory,
        delete_scope=delete_scope,
        metadata=metadata,
        metadata_merger=metadata_merger,
    )


@default_docs
@normalize_args
def store_delayed_as_dataset(
    delayed_tasks: list[Delayed],
    store,
    dataset_uuid=None,
    metadata=None,
    df_serializer=None,
    overwrite=False,
    metadata_merger=None,
    metadata_version=naming.DEFAULT_METADATA_VERSION,
    partition_on=None,
    metadata_storage_format=naming.DEFAULT_METADATA_STORAGE_FORMAT,
    table_name: str = SINGLE_TABLE,
    secondary_indices=None,
) -> Delayed:
    """Transform and store a list of dictionaries containing dataframes to a
    plateau dataset in store.

    Parameters
    ----------
    """
    store = lazy_store(store)
    if dataset_uuid is None:
        dataset_uuid = gen_uuid()

    if not overwrite:
        raise_if_dataset_exists(dataset_uuid=dataset_uuid, store=store)

    raise_if_indices_overlap(partition_on, secondary_indices)

    input_to_mps = partial(
        parse_input_to_metapartition,
        metadata_version=metadata_version,
        table_name=table_name,
    )
    mps = map_delayed(input_to_mps, delayed_tasks)

    if partition_on:
        mps = map_delayed(MetaPartition.partition_on, mps, partition_on=partition_on)

    if secondary_indices:
        mps = map_delayed(MetaPartition.build_indices, mps, columns=secondary_indices)

    mps = map_delayed(
        MetaPartition.store_dataframes,
        mps,
        store=store,
        df_serializer=df_serializer,
        dataset_uuid=dataset_uuid,
    )

    return delayed(store_dataset_from_partitions)(
        list(mps),
        dataset_uuid=dataset_uuid,
        store=store,
        dataset_metadata=metadata,
        metadata_merger=metadata_merger,
        metadata_storage_format=metadata_storage_format,
    )



"""This module is a collection of helper functions."""

import collections
import inspect
import logging
from collections.abc import Iterable
from typing import Literal, cast, overload

import decorator
import pandas as pd

from plateau.core.dataset import DatasetMetadata, DatasetMetadataBase
from plateau.core.factory import _ensure_factory
from plateau.core.typing import StoreFactory, StoreInput
from plateau.core.utils import ensure_store, lazy_store

signature = inspect.signature


LOGGER = logging.getLogger(__name__)


class InvalidObject:
    """Sentinel to mark keys for removal."""

    pass


def combine_metadata(dataset_metadata: list[dict], append_to_list: bool = True) -> dict:
    """Merge a list of dictionaries.

    The merge is performed in such a way, that only keys which
    are present in **all** dictionaries are kept in the final result.

    If lists are encountered, the values of the result will be the
    concatenation of all list values in the order of the supplied dictionary list.
    This behaviour may be changed by using append_to_list

    Parameters
    ----------
    dataset_metadata
        The list of dictionaries (usually metadata) to be combined.
    append_to_list
        If True, all values are concatenated. If False, only unique values are kept
    """
    meta = _combine_metadata(dataset_metadata, append_to_list)
    return _remove_invalids(meta)


def _remove_invalids(dct):
    if not isinstance(dct, dict):
        return {}

    new_dict = {}
    for key, value in dct.items():
        if isinstance(value, dict):
            tmp = _remove_invalids(value)
            # Do not propagate empty dicts
            if tmp:
                new_dict[key] = tmp
        elif not isinstance(value, InvalidObject):
            new_dict[key] = value
    return new_dict


def _combine_metadata(dataset_metadata, append_to_list):
    assert isinstance(dataset_metadata, list)
    if len(dataset_metadata) == 1:
        return dataset_metadata.pop()

    # In case the input list has only two elements, we can do simple comparison
    if len(dataset_metadata) > 2:
        first = _combine_metadata(dataset_metadata[::2], append_to_list)
        second = _combine_metadata(dataset_metadata[1::2], append_to_list)
        final = _combine_metadata([first, second], append_to_list)
        return final
    else:
        first = dataset_metadata.pop()
        second = dataset_metadata.pop()
        if first == second:
            return first
        # None is harmless and may occur if a key appears in one but not the other dict
        elif first is None or second is None:
            return first if first is not None else second
        elif isinstance(first, dict) and isinstance(second, dict):
            new_dict = {}
            keys = set(first.keys())
            keys.update(second.keys())
            for key in keys:
                new_dict[key] = _combine_metadata(
                    [first.get(key), second.get(key)], append_to_list
                )
            return new_dict
        elif isinstance(first, list) and isinstance(second, list):
            new_list = first + second
            if append_to_list:
                return new_list
            else:
                return list(set(new_list))
        else:
            return InvalidObject()


def _ensure_compatible_indices(
    dataset: DatasetMetadataBase | None,
    secondary_indices: Iterable[str],
) -> list[str]:
    if dataset:
        ds_secondary_indices = sorted(dataset.secondary_indices.keys())

        if secondary_indices and not set(secondary_indices).issubset(
            ds_secondary_indices
        ):
            raise ValueError(
                f"Incorrect indices provided for dataset.\n"
                f"Expected: {ds_secondary_indices}\n"
                f"But got: {secondary_indices}"
            )

        return ds_secondary_indices
    return sorted(secondary_indices)


def validate_partition_keys(
    dataset_uuid,
    store,
    ds_factory,
    default_metadata_version,
    partition_on,
):
    if ds_factory or DatasetMetadata.exists(dataset_uuid, ensure_store(store)):
        ds_factory = _ensure_factory(
            dataset_uuid=dataset_uuid,
            store=store,
            factory=ds_factory,
        )

        ds_metadata_version = ds_factory.metadata_version
        if partition_on:
            if not isinstance(partition_on, list):
                partition_on = [partition_on]
            if partition_on != ds_factory.partition_keys:
                raise ValueError(
                    "Incompatible set of partition keys encountered. "
                    f"Input partitioning was `{partition_on}` while actual dataset was `{ds_factory.partition_keys}`"
                )
        else:
            partition_on = ds_factory.partition_keys
    else:
        ds_factory = None
        ds_metadata_version = default_metadata_version
    return ds_factory, ds_metadata_version, partition_on


_NORMALIZE_ARGS_LIST = [
    "partition_on",
    "delete_scope",
    "secondary_indices",
    "sort_partitions_by",
    "bucket_by",
]

_NORMALIZE_ARGS = _NORMALIZE_ARGS_LIST + ["store", "dispatch_by"]
_NormalizeArgsLiteral = Literal[
    "partition_on",
    "delete_scope",
    "secondary_indices",
    "sort_partitions_by",
    "bucket_by",
    "store",
    "dispatch_by",
]


@overload
def normalize_arg(
    arg_name: Literal[
        "partition_on",
        "delete_scope",
        "secondary_indices",
        "bucket_by",
        "sort_partitions_by",
        "dispatch_by",
    ],
    old_value: None,
) -> None: ...


@overload
def normalize_arg(
    arg_name: Literal[
        "partition_on",
        "delete_scope",
        "secondary_indices",
        "bucket_by",
        "sort_partitions_by",
        "dispatch_by",
    ],
    old_value: str | list[str],
) -> list[str]: ...


@overload
def normalize_arg(
    arg_name: Literal["store"], old_value: StoreInput | None
) -> StoreFactory: ...


def normalize_arg(arg_name, old_value):
    """Normalizes an argument according to pre-defined types.

    Type A:

    * "partition_on"
    * "delete_scope"
    * "secondary_indices"
    * "dispatch_by"

    will be converted to a list. If it is None, an empty list will be created

    Type B:
    * "store"

    Will be converted to a callable returning

    :meta private:
    """

    def _make_list(_args):
        if isinstance(_args, str | bytes | int | float):
            return [_args]
        if _args is None:
            return []
        if isinstance(_args, set | frozenset | dict):
            raise ValueError(f"{type(_args)} is incompatible for normalisation.")
        return list(_args)

    if arg_name in _NORMALIZE_ARGS_LIST:
        if old_value is None:
            return []
        elif isinstance(old_value, list):
            return old_value
        else:
            return _make_list(old_value)
    elif arg_name == "dispatch_by":
        if old_value is None:
            return old_value
        elif isinstance(old_value, list):
            return old_value
        else:
            return _make_list(old_value)
    elif arg_name == "store" and old_value is not None:
        return lazy_store(old_value)

    return old_value


@decorator.decorator
def normalize_args(function, *args, **kwargs):
    sig = signature(function)

    def _wrapper(*args, **kwargs):
        for arg_name in _NORMALIZE_ARGS:
            arg_name = cast(_NormalizeArgsLiteral, arg_name)
            if arg_name in sig.parameters.keys():
                ix = inspect.getfullargspec(function).args.index(arg_name)
                if arg_name in kwargs:
                    kwargs[arg_name] = normalize_arg(arg_name, kwargs[arg_name])
                elif len(args) > ix:
                    new_args = list(args)
                    new_args[ix] = normalize_arg(arg_name, args[ix])
                    args = tuple(new_args)
                else:
                    kwargs[arg_name] = normalize_arg(arg_name, None)
        return function(*args, **kwargs)

    return _wrapper(*args, **kwargs)


def extract_duplicates(lst):
    """Return all items of a list that occur more than once.

    Parameters
    ----------
    lst: List[Any]

    Returns
    -------
    lst: List[Any]
    """

    return [item for item, count in collections.Counter(lst).items() if count > 1]


def align_categories(dfs, categoricals):
    """Takes a list of dataframes with categorical columns and determines the
    superset of categories. All specified columns will then be cast to the same
    `pd.CategoricalDtype`

    Parameters
    ----------
    dfs: List[pd.DataFrame]
        A list of dataframes for which the categoricals should be aligned
    categoricals: List[str]
        Columns holding categoricals which should be aligned

    Returns
    -------
    List[pd.DataFrame]
        A list with aligned dataframes
    """
    if len(categoricals) == 0:
        return dfs

    col_dtype = {}

    for column in categoricals:
        position_largest_df = None
        categories = set()
        largest_df_categories = set()
        for ix, df in enumerate(dfs):
            ser = df[column]
            if not isinstance(ser.dtype, pd.CategoricalDtype):
                cats = ser.dropna().unique()
                LOGGER.info(
                    "Encountered non-categorical type where categorical was expected\n"
                    f"Found at index position {ix} for column {column}\n"
                    f"Dtypes: {df.dtypes}"
                )
            else:
                cats = ser.cat.categories
                length = len(df)
                if position_largest_df is None or length > position_largest_df[0]:
                    position_largest_df = (length, ix)
                if position_largest_df[1] == ix:
                    largest_df_categories = cats
            categories.update(cats)

        # use the categories of the largest DF as a baseline to avoid having
        # to rewrite its codes. Append the remainder and sort it for reproducibility
        categories_lst = list(largest_df_categories) + sorted(
            set(categories) - set(largest_df_categories)
        )
        cat_dtype = pd.api.types.CategoricalDtype(categories_lst, ordered=False)
        col_dtype[column] = cat_dtype

    return_dfs = []
    for df in dfs:
        try:
            new_df = df.astype(col_dtype, copy=False)
        except ValueError as verr:
            cat_types = {
                col: dtype.categories.dtype for col, dtype in col_dtype.items()
            }
            # Should be fixed by pandas>=0.24.0
            if "buffer source array is read-only" in str(verr):
                new_df = df.astype(cat_types)
                new_df = new_df.astype(col_dtype)
            else:
                raise verr
        return_dfs.append(new_df)
    return return_dfs


def sort_values_categorical(df: pd.DataFrame, columns: list[str] | str) -> pd.DataFrame:
    """Sort a dataframe lexicographically by the categories of column
    `column`"""
    if not isinstance(columns, list):
        columns = [columns]
    for col in columns:
        if isinstance(df[col].dtype, pd.CategoricalDtype):
            cat_accessor = df[col].cat
            df[col] = cat_accessor.reorder_categories(
                sorted(cat_accessor.categories), ordered=True
            )
    return df.sort_values(by=columns).reset_index(drop=True)


def raise_if_indices_overlap(partition_on, secondary_indices):
    partition_secondary_overlap = set(partition_on) & set(secondary_indices)
    if partition_secondary_overlap:
        raise RuntimeError(
            f"Cannot create secondary index on partition columns: {partition_secondary_overlap}"
        )
